/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.info.ApplicationLocale;
import org.dussan.vaadin.dvirtual.info.KeyboardLayout;
import org.dussan.vaadin.dvirtual.info.TranslationLocale;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class DVirtualConfig
    implements Serializable
{

    private static final long serialVersionUID = -6081303135795348774L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int FIRST_ELEMENT = 0;

    private static final int NO_ELEMENTS = 0;

    private static final String ADMIN_CONFIG = "admin_config";

    private static final String ALL_RUNNING_IN_SAME_INSTANCE = "all_running_in_same_instance";

    private static final String APPLICATION_CONFIG = "application_config";

    private static final String APPLICATION_LOCALE = "application_locale";

    private static final String APPLICATIONS_BAR_VISIBLE = "applications_bar_visible";

    private static final String APPLICATIONS_GROUP_CONFIG = "applications_group_config";

    private static final String DATA_VERSION = "version";

    private static final String DATA_VERSION_1 = "1";

    private static final String DATABASE_CONFIG = "database_config";

    private static final String ERROR_CANNOT_LOAD = "Cannot load dVirtual configuration, cause: {}";

    private static final String KEYBOARD_LAYOUT = "keyboard_layout";

    private static final String LOGIN_TITLE = "login_title";

    private static final String MOUSE_MOVE_THROTTLE_PERIOD = "mouse_move_throttle_period";

    private static final String SALT = "salt";

    private static final String SALT_IV = "salt_iv";

    private static final String SALT_KEY = "salt_key";

    private static final String SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES = "http://xml.org/sax/features/external-general-entities";

    private static final String SERVER_CONFIG = "server_config";

    private static final String SERVERS_GROUP_CONFIG = "servers_group_config";

    private static final String SESSION_TIMEOUT = "session_timeout";

    private static final String STARTUP_NOTIFICATION_VISIBLE = "startup_notifications_visible";

    private static final String TOOLTIP_VISIBLE = "tooltip_visible";

    private static final String TRANSLATION_LOCALE = "translation_locale";

    private static final String USER_CONFIG = "user_config";

    private Boolean allRunningInSameInstance = null;

    private Boolean applicationsBarVisible = null;

    private Boolean startupNotificationsVisible = null;

    private Boolean tooltipsVisible = null;

    private Integer mouseMoveThrottlePeriod = null;

    private Integer sessionTimeout = null;

    private String adminConfig = null;

    private String applicationConfig = null;

    private String applicationsGroupConfig = null;

    private String bandwidthLimit = null;

    private String databaseConfig = null;

    private String loginTitle = null;

    private String saltIV = null;

    private String saltKey = null;

    private String saltValue = null;

    private String serverConfig = null;

    private String serversGroupConfig = null;

    private String userConfig = null;

    private ApplicationLocale applicationLocale = null;

    private KeyboardLayout keyboardLayout = null;

    private TranslationLocale translationLocale = null;

    public DVirtualConfig( String configFolder, String configFile )
    {
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newDefaultInstance();
            documentBuilderFactory.setFeature( SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES, false );
            documentBuilderFactory.setValidating( false );
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document doc = documentBuilder.parse( new InputSource( fileReader ) );
            Element rootNode = doc.getDocumentElement();

            // get data version
            String dataVersion = null;
            for ( int index = 0; index < rootNode.getAttributes()
                                                 .getLength(); index++ )
            {
                Node node = rootNode.getAttributes()
                                    .item( index );
                if ( Objects.nonNull( node ) && StringUtils.equals( node.getNodeName(), DATA_VERSION ) && StringUtils.isNotBlank( node.getTextContent() ) )
                {
                    dataVersion = StringUtils.stripToEmpty( node.getTextContent() );
                }
            }

            if ( StringUtils.equals( dataVersion, DATA_VERSION_1 ) )
            {
                // salt
                if ( rootNode.getElementsByTagName( SALT )
                             .getLength() != NO_ELEMENTS )
                {
                    setSalt( StringUtils.stripToEmpty( rootNode.getElementsByTagName( SALT )
                                                               .item( FIRST_ELEMENT )
                                                               .getTextContent() ) );
                }

                // salt IV
                if ( rootNode.getElementsByTagName( SALT_IV )
                             .getLength() != NO_ELEMENTS )
                {
                    setSaltIV( StringUtils.stripToEmpty( rootNode.getElementsByTagName( SALT_IV )
                                                                 .item( FIRST_ELEMENT )
                                                                 .getTextContent() ) );
                }

                // salt key
                if ( rootNode.getElementsByTagName( SALT_KEY )
                             .getLength() != NO_ELEMENTS )
                {
                    setSaltKey( StringUtils.stripToEmpty( rootNode.getElementsByTagName( SALT_KEY )
                                                                  .item( FIRST_ELEMENT )
                                                                  .getTextContent() ) );
                }

                // all running in same instance
                if ( rootNode.getElementsByTagName( ALL_RUNNING_IN_SAME_INSTANCE )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( ALL_RUNNING_IN_SAME_INSTANCE )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setAllRunningInSameInstance( StringUtils.isNotBlank( value ) ? Boolean.parseBoolean( value ) : null );
                }

                // applications bar visible
                if ( rootNode.getElementsByTagName( APPLICATIONS_BAR_VISIBLE )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( APPLICATIONS_BAR_VISIBLE )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setApplicationsBarVisible( StringUtils.isNotBlank( value ) ? Boolean.parseBoolean( value ) : null );
                }

                // startup notifications visible
                if ( rootNode.getElementsByTagName( STARTUP_NOTIFICATION_VISIBLE )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( STARTUP_NOTIFICATION_VISIBLE )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setStartupNotificationsVisible( StringUtils.isNotBlank( value ) ? Boolean.parseBoolean( value ) : null );
                }

                // tooltip visible
                if ( rootNode.getElementsByTagName( TOOLTIP_VISIBLE )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( TOOLTIP_VISIBLE )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setTooltipsVisible( StringUtils.isNotBlank( value ) ? Boolean.parseBoolean( value ) : null );
                }

                // mouse move throttle period
                if ( rootNode.getElementsByTagName( MOUSE_MOVE_THROTTLE_PERIOD )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( MOUSE_MOVE_THROTTLE_PERIOD )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setMouseMoveThrottlePeriod( StringUtils.isNotBlank( value ) ? Integer.parseInt( value ) : null );
                }

                // session timeout
                if ( rootNode.getElementsByTagName( SESSION_TIMEOUT )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( SESSION_TIMEOUT )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setSessionTimeout( StringUtils.isNotBlank( value ) ? Integer.parseInt( value ) : null );
                }

                // admin config
                if ( rootNode.getElementsByTagName( ADMIN_CONFIG )
                             .getLength() != NO_ELEMENTS )
                {
                    setAdminConfig( StringUtils.stripToEmpty( rootNode.getElementsByTagName( ADMIN_CONFIG )
                                                                      .item( FIRST_ELEMENT )
                                                                      .getTextContent() ) );
                }

                // application config
                if ( rootNode.getElementsByTagName( APPLICATION_CONFIG )
                             .getLength() != NO_ELEMENTS )
                {
                    setApplicationConfig( StringUtils.stripToEmpty( rootNode.getElementsByTagName( APPLICATION_CONFIG )
                                                                            .item( FIRST_ELEMENT )
                                                                            .getTextContent() ) );
                }

                // applications group config
                if ( rootNode.getElementsByTagName( APPLICATIONS_GROUP_CONFIG )
                             .getLength() != NO_ELEMENTS )
                {
                    setApplicationsGroupConfig( StringUtils.stripToEmpty( rootNode.getElementsByTagName( APPLICATIONS_GROUP_CONFIG )
                                                                                  .item( FIRST_ELEMENT )
                                                                                  .getTextContent() ) );
                }

                // database config
                if ( rootNode.getElementsByTagName( DATABASE_CONFIG )
                             .getLength() != NO_ELEMENTS )
                {
                    setDatabaseConfig( StringUtils.stripToEmpty( rootNode.getElementsByTagName( DATABASE_CONFIG )
                                                                         .item( FIRST_ELEMENT )
                                                                         .getTextContent() ) );
                }

                // login title
                if ( rootNode.getElementsByTagName( LOGIN_TITLE )
                             .getLength() != NO_ELEMENTS )
                {
                    setLoginTitle( StringUtils.stripToEmpty( rootNode.getElementsByTagName( LOGIN_TITLE )
                                                                     .item( FIRST_ELEMENT )
                                                                     .getTextContent() ) );
                }

                // server config
                if ( rootNode.getElementsByTagName( SERVER_CONFIG )
                             .getLength() != NO_ELEMENTS )
                {
                    setServerConfig( StringUtils.stripToEmpty( rootNode.getElementsByTagName( SERVER_CONFIG )
                                                                       .item( FIRST_ELEMENT )
                                                                       .getTextContent() ) );
                }

                // servers group config
                if ( rootNode.getElementsByTagName( SERVERS_GROUP_CONFIG )
                             .getLength() != NO_ELEMENTS )
                {
                    setServersGroupConfig( StringUtils.stripToEmpty( rootNode.getElementsByTagName( SERVERS_GROUP_CONFIG )
                                                                             .item( FIRST_ELEMENT )
                                                                             .getTextContent() ) );
                }

                // user config
                if ( rootNode.getElementsByTagName( USER_CONFIG )
                             .getLength() != NO_ELEMENTS )
                {
                    setUserConfig( StringUtils.stripToEmpty( rootNode.getElementsByTagName( USER_CONFIG )
                                                                     .item( FIRST_ELEMENT )
                                                                     .getTextContent() ) );
                }

                // application locale
                if ( rootNode.getElementsByTagName( APPLICATION_LOCALE )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( APPLICATION_LOCALE )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setApplicationLocale( StringUtils.isNotBlank( value ) ? ApplicationLocale.getApplicationLocale( value ) : null );
                }

                // keyboard layout
                if ( rootNode.getElementsByTagName( KEYBOARD_LAYOUT )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( KEYBOARD_LAYOUT )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setKeyboardLayout( StringUtils.isNotBlank( value ) ? KeyboardLayout.getKeyboardLayout( value ) : null );
                }

                // translation locale
                if ( rootNode.getElementsByTagName( TRANSLATION_LOCALE )
                             .getLength() != NO_ELEMENTS )
                {
                    String value = rootNode.getElementsByTagName( TRANSLATION_LOCALE )
                                           .item( FIRST_ELEMENT )
                                           .getTextContent();
                    setTranslationLocale( StringUtils.isNotBlank( value ) ? TranslationLocale.getTranslationLocale( value ) : null );
                }
            }
        }
        catch ( IOException | SAXException | ParserConfigurationException exception )
        {
            LOGGER.error( ERROR_CANNOT_LOAD, exception.getMessage() );
        }
    }

    public boolean isInitialized()
    {
        return ( StringUtils.isNotBlank( getSalt() ) && StringUtils.isNotBlank( getSaltIV() ) && StringUtils.isNotBlank( getSaltKey() ) );
    }

    public String getSalt()
    {
        return saltValue;
    }

    public void setSalt( String salt )
    {
        this.saltValue = salt;
    }

    public String getSaltIV()
    {
        return saltIV;
    }

    public void setSaltIV( String saltIV )
    {
        this.saltIV = saltIV;
    }

    public String getSaltKey()
    {
        return saltKey;
    }

    public void setSaltKey( String saltKey )
    {
        this.saltKey = saltKey;
    }

    public String getBandwidthLimit()
    {
        return bandwidthLimit;
    }

    public void setBandwidthLimit( String bandwidthLimit )
    {
        this.bandwidthLimit = bandwidthLimit;
    }

    public Boolean isAllRunningInSameInstance()
    {
        return allRunningInSameInstance;
    }

    public void setAllRunningInSameInstance( Boolean allRunningInSameInstance )
    {
        this.allRunningInSameInstance = allRunningInSameInstance;
    }

    public Boolean isApplicationsBarVisible()
    {
        return applicationsBarVisible;
    }

    public void setApplicationsBarVisible( Boolean applicationsBarVisible )
    {
        this.applicationsBarVisible = applicationsBarVisible;
    }

    public Boolean isStartupNotificationsVisible()
    {
        return startupNotificationsVisible;
    }

    public void setStartupNotificationsVisible( Boolean startupNotificationsVisible )
    {
        this.startupNotificationsVisible = startupNotificationsVisible;
    }

    public Boolean isTooltipsVisible()
    {
        return tooltipsVisible;
    }

    public void setTooltipsVisible( Boolean tooltipsVisible )
    {
        this.tooltipsVisible = tooltipsVisible;
    }

    public Integer getMouseMoveThrottlePeriod()
    {
        return mouseMoveThrottlePeriod;
    }

    public void setMouseMoveThrottlePeriod( Integer mouseMoveThrottlePeriod )
    {
        this.mouseMoveThrottlePeriod = mouseMoveThrottlePeriod;
    }

    public Integer getSessionTimeout()
    {
        return sessionTimeout;
    }

    public void setSessionTimeout( Integer sessionTimeout )
    {
        this.sessionTimeout = sessionTimeout;
    }

    public String getAdminConfig()
    {
        return adminConfig;
    }

    public void setAdminConfig( String adminConfig )
    {
        this.adminConfig = adminConfig;
    }

    public String getApplicationConfig()
    {
        return applicationConfig;
    }

    public void setApplicationConfig( String applicationConfig )
    {
        this.applicationConfig = applicationConfig;
    }

    public String getApplicationsGroupConfig()
    {
        return applicationsGroupConfig;
    }

    public void setApplicationsGroupConfig( String applicationsGroupConfig )
    {
        this.applicationsGroupConfig = applicationsGroupConfig;
    }

    public String getDatabaseConfig()
    {
        return databaseConfig;
    }

    public void setDatabaseConfig( String databaseConfig )
    {
        this.databaseConfig = databaseConfig;
    }

    public String getLoginTitle()
    {
        return loginTitle;
    }

    public void setLoginTitle( String loginTitle )
    {
        this.loginTitle = loginTitle;
    }

    public String getServerConfig()
    {
        return serverConfig;
    }

    public void setServerConfig( String serverConfig )
    {
        this.serverConfig = serverConfig;
    }

    public String getServersGroupConfig()
    {
        return serversGroupConfig;
    }

    public void setServersGroupConfig( String serversGroupConfig )
    {
        this.serversGroupConfig = serversGroupConfig;
    }

    public String getUserConfig()
    {
        return userConfig;
    }

    public void setUserConfig( String userConfig )
    {
        this.userConfig = userConfig;
    }

    public ApplicationLocale getApplicationLocale()
    {
        return applicationLocale;
    }

    public void setApplicationLocale( ApplicationLocale applicationLocale )
    {
        this.applicationLocale = applicationLocale;
    }

    public KeyboardLayout getKeyboardLayout()
    {
        return keyboardLayout;
    }

    public void setKeyboardLayout( KeyboardLayout keyboardLayout )
    {
        this.keyboardLayout = keyboardLayout;
    }

    public TranslationLocale getTranslationLocale()
    {
        return translationLocale;
    }

    public void setTranslationLocale( TranslationLocale translationLocale )
    {
        this.translationLocale = translationLocale;
    }

}
