/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang3.StringUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class DVirtualSetupConfig
    implements Serializable
{

    private static final long serialVersionUID = 8083994766554353176L;

    private static final int FIRST_ELEMENT = 0;

    private static final int NO_ELEMENTS = 0;

    private static final String PAZZWORD = "password";

    private static final String SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES = "http://xml.org/sax/features/external-general-entities";

    private static final String UZERNAME = "username";

    private static final String DATA_VERSION = "version";

    private static final String DATA_VERSION_1 = "1";

    private String password = null;

    private String username = null;

    public DVirtualSetupConfig( String configFolder, String configFile )
    {
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newDefaultInstance();
            documentBuilderFactory.setFeature( SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES, false );
            documentBuilderFactory.setValidating( false );
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document doc = documentBuilder.parse( new InputSource( fileReader ) );
            Element rootNode = doc.getDocumentElement();

            // get data version
            String dataVersion = null;
            for ( int index = 0; index < rootNode.getAttributes()
                                                 .getLength(); index++ )
            {
                Node node = rootNode.getAttributes()
                                    .item( index );
                if ( Objects.nonNull( node ) && StringUtils.equals( node.getNodeName(), DATA_VERSION ) && StringUtils.isNotBlank( node.getTextContent() ) )
                {
                    dataVersion = StringUtils.stripToEmpty( node.getTextContent() );
                }
            }

            if ( StringUtils.equals( dataVersion, DATA_VERSION_1 ) )
            {
                // password
                if ( rootNode.getElementsByTagName( PAZZWORD )
                             .getLength() != NO_ELEMENTS )
                {
                    setPassword( StringUtils.stripToNull( rootNode.getElementsByTagName( PAZZWORD )
                                                                  .item( FIRST_ELEMENT )
                                                                  .getTextContent() ) );
                }

                // user
                if ( rootNode.getElementsByTagName( UZERNAME )
                             .getLength() != NO_ELEMENTS )
                {
                    setUsername( StringUtils.stripToNull( rootNode.getElementsByTagName( UZERNAME )
                                                                  .item( FIRST_ELEMENT )
                                                                  .getTextContent() ) );
                }
            }
        }
        catch ( IOException | SAXException | ParserConfigurationException exception )
        {
            reset();
        }
    }

    public boolean isInitialized()
    {
        return ( StringUtils.isNotBlank( password ) && StringUtils.isNotBlank( username ) );
    }

    public void reset()
    {
        setPassword( null );
        setUsername( null );
    }

    public String getPassword()
    {
        return password;
    }

    public void setPassword( String password )
    {
        this.password = password;
    }

    public String getUsername()
    {
        return username;
    }

    public void setUsername( String username )
    {
        this.username = username;
    }

}
