/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class ServersGroupConfig
    implements Serializable
{

    private static final long serialVersionUID = -26967897442353552L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final String DATA_VERSION = "version";

    private static final String DATA_VERSION_1 = "1";

    private static final String ERROR_CANNOT_LOAD = "Cannot load servers group configuration, cause: {}";

    private static final String SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES = "http://xml.org/sax/features/external-general-entities";

    private static final String SERVERS_GROUP = "servers_group";

    private static final String SERVERS_GROUP_ENABLED = "enabled";

    private static final String SERVERS_GROUP_ID = "id";

    private static final String SERVERS_GROUP_NAME = "name";

    private static final String SERVERS_GROUP_NOTE = "note";

    private static final String SERVERS_GROUP_SERVER_ID = "server_id";

    private static final String SERVERS_GROUP_SERVER_IDS = "server_ids";

    private static final String SERVERS_GROUP_SERVER_NAME = "server_name";

    private static final String SERVERS_GROUP_SERVER_NAMES = "server_names";

    private static final String SPLITTER_COMMAS = ",";

    private static final String SPLITTER_SPACE = " ";

    private transient List<ServersGroupData> serversGroupDatas = null;

    private ServersGroupConfig()
    {
        serversGroupDatas = new ArrayList<>();
    }

    public ServersGroupConfig( String configFolder, String configFile )
    {
        this();
        try (FileReader fileReader = new FileReader( Paths.get( configFolder, configFile )
                                                          .toFile() ))
        {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newDefaultInstance();
            documentBuilderFactory.setFeature( SAX_FEATURES_EXTERNAL_GENERAL_ENTITIES, false );
            documentBuilderFactory.setValidating( false );
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
            Document document = documentBuilder.parse( new InputSource( fileReader ) );
            Element rootNode = document.getDocumentElement();

            // get data version
            String dataVersion = null;
            for ( int index = 0; index < rootNode.getAttributes()
                                                 .getLength(); index++ )
            {
                Node node = rootNode.getAttributes()
                                    .item( index );
                if ( Objects.nonNull( node ) && StringUtils.equals( node.getNodeName(), DATA_VERSION ) && StringUtils.isNotBlank( node.getTextContent() ) )
                {
                    dataVersion = StringUtils.stripToEmpty( node.getTextContent() );
                }
            }

            // servers group data
            if ( StringUtils.equals( dataVersion, DATA_VERSION_1 ) )
            {
                NodeList nodeList = rootNode.getElementsByTagName( SERVERS_GROUP );
                for ( int index = 0; index < nodeList.getLength(); index++ )
                {
                    Node serversGroupNode = nodeList.item( index );
                    if ( serversGroupNode.getNodeType() == Node.ELEMENT_NODE && serversGroupNode.hasChildNodes() )
                    {
                        ServersGroupData serversGroupData = new ServersGroupData();
                        serversGroupData.setId( ServersGroupData.NEW_SERVERS_GROUP );
                        serversGroupData.setServerDatas( new ArrayList<>() );

                        NodeList serversGroupNodeList = serversGroupNode.getChildNodes();
                        for ( int i = 0; i < serversGroupNodeList.getLength(); i++ )
                        {
                            Node serversGroupChildNode = serversGroupNodeList.item( i );
                            String serversGroupNodeText = StringUtils.stripToEmpty( serversGroupChildNode.getTextContent() );
                            switch ( serversGroupChildNode.getNodeName() )
                            {
                                case SERVERS_GROUP_ENABLED:
                                    serversGroupData.setEnabled( Boolean.parseBoolean( serversGroupNodeText ) );
                                    break;
                                case SERVERS_GROUP_ID:
                                    if ( StringUtils.isNumeric( serversGroupNodeText ) )
                                    {
                                        serversGroupData.setId( Math.abs( Long.parseLong( serversGroupNodeText ) ) );
                                    }
                                    break;
                                case SERVERS_GROUP_NAME:
                                    serversGroupData.setName( serversGroupNodeText );
                                    break;
                                case SERVERS_GROUP_NOTE:
                                    serversGroupData.setNote( serversGroupNodeText );
                                    break;
                                case SERVERS_GROUP_SERVER_ID:
                                    if ( StringUtils.isNumeric( serversGroupNodeText ) )
                                    {
                                        List<ServerData> datas = serversGroupData.getServerDatas();
                                        ServerData serverData = new ServerData();
                                        serverData.setId( Math.abs( Long.parseLong( serversGroupNodeText ) ) );
                                        serversGroupData.setServerDatas( datas );
                                    }
                                    break;
                                case SERVERS_GROUP_SERVER_IDS:
                                    if ( StringUtils.isNumericSpace( serversGroupNodeText.replace( SPLITTER_COMMAS, SPLITTER_SPACE ) ) )
                                    {
                                        List<ServerData> datas = serversGroupData.getServerDatas();
                                        Arrays.asList( StringUtils.split( StringUtils.replace( serversGroupNodeText, SPLITTER_SPACE, SPLITTER_COMMAS ), SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNumeric( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ServerData serverData = new ServerData();
                                                  serverData.setId( Math.abs( Long.parseLong( StringUtils.strip( data ) ) ) );
                                                  datas.add( serverData );
                                              } );
                                        serversGroupData.setServerDatas( datas );
                                    }
                                    break;
                                case SERVERS_GROUP_SERVER_NAME:
                                    if ( StringUtils.isNotBlank( serversGroupNodeText ) )
                                    {
                                        List<ServerData> datas = serversGroupData.getServerDatas();
                                        ServerData serverData = new ServerData();
                                        serverData.setName( serversGroupNodeText );
                                        datas.add( serverData );
                                        serversGroupData.setServerDatas( datas );
                                    }
                                    break;
                                case SERVERS_GROUP_SERVER_NAMES:
                                    if ( StringUtils.isNotBlank( serversGroupNodeText ) )
                                    {
                                        List<ServerData> datas = serversGroupData.getServerDatas();
                                        Arrays.asList( StringUtils.split( serversGroupNodeText, SPLITTER_COMMAS ) )
                                              .stream()
                                              .filter( data -> StringUtils.isNotEmpty( StringUtils.stripToEmpty( data ) ) )
                                              .forEach( data -> {
                                                  ServerData serverData = new ServerData();
                                                  serverData.setName( StringUtils.strip( data ) );
                                                  datas.add( serverData );
                                              } );
                                        serversGroupData.setServerDatas( datas );
                                    }
                                    break;
                                default:
                                    // nothing to do
                                    break;
                            }
                        }

                        // add new servers group data
                        if ( StringUtils.isNotBlank( serversGroupData.getName() ) && CollectionUtils.isNotEmpty( serversGroupData.getServerDatas() ) )
                        {
                            serversGroupDatas.add( serversGroupData );
                        }
                    }
                }
            }
        }
        catch ( IOException | SAXException | ParserConfigurationException exception )
        {
            LOGGER.info( ERROR_CANNOT_LOAD, exception.getMessage() );
        }
    }

    public List<ServersGroupData> getServersGroupDatas()
    {
        return serversGroupDatas;
    }

}
