/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.database;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaBuilder.In;
import javax.persistence.criteria.CriteriaDelete;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Path;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationColumns;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationsGroupColumns;
import org.dussan.vaadin.dvirtual.database.columns.ServerColumns;
import org.dussan.vaadin.dvirtual.database.columns.ServersGroupColumns;
import org.dussan.vaadin.dvirtual.database.columns.UserColumns;
import org.dussan.vaadin.dvirtual.database.columns.UserSessionColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData_;
import org.dussan.vaadin.dvirtual.database.data.ApplicationServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationServersGroupData_;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupApplicationData_;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData_;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.database.data.IconData_;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServerData_;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData_;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupServerData_;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationData_;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationsGroupData_;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.database.data.UserData_;
import org.dussan.vaadin.dvirtual.database.data.UserSessionApplicationData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionApplicationData_;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData_;
import org.dussan.vaadin.dvirtual.database.filters.ApplicationFilter;
import org.dussan.vaadin.dvirtual.database.filters.ApplicationsGroupFilter;
import org.dussan.vaadin.dvirtual.database.filters.ServerFilter;
import org.dussan.vaadin.dvirtual.database.filters.ServersGroupFilter;
import org.dussan.vaadin.dvirtual.database.filters.UserFilter;
import org.dussan.vaadin.dvirtual.database.filters.UserSessionFilter;
import org.dussan.vaadin.dvirtual.database.orders.ApplicationOrder;
import org.dussan.vaadin.dvirtual.database.orders.ApplicationsGroupOrder;
import org.dussan.vaadin.dvirtual.database.orders.ServerOrder;
import org.dussan.vaadin.dvirtual.database.orders.ServersGroupOrder;
import org.dussan.vaadin.dvirtual.database.orders.UserOrder;
import org.dussan.vaadin.dvirtual.utils.HibernateUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.slf4j.Logger;

import com.vaadin.flow.component.icon.VaadinIcon;

public class Database
    implements Serializable
{

    private static final long serialVersionUID = 7245982258980972923L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    public static final String ORDER_DESCENDING = "-";

    public static final String PROPERTIES = "properties";

    private static final int FIRST_ELEMENT = 0;

    private static final int ONE_ELEMENT = 1;

    private static final long NEW_ELEMENT = 0;

    private static final long NO_ELEMENTS = 0;

    private static final String ATTRIBUTE_ICON = "icon";

    private static final VaadinIcon DEFAULT_ICON = VaadinIcon.AREA_SELECT;

    private boolean initialized = false;

    private HibernateUtil hibernateUtil = null;

    public Database( String configFolder, String configFile )
    {
        // Hibernate initialization
        hibernateUtil = new HibernateUtil( configFolder, configFile );
        initialized = hibernateUtil.isInitialized();
    }

    private List<Long> getValues( Long... values )
    {
        List<Long> longs = new ArrayList<>();
        if ( ArrayUtils.isNotEmpty( values ) )
        {
            longs.addAll( Arrays.asList( values ) );
            longs.removeIf( Objects::isNull );
        }
        return ( new ArrayList<>( new LinkedHashSet<>( longs ) ) );
    }

    private List<String> getValues( String... values )
    {
        List<String> strings = new ArrayList<>();
        if ( ArrayUtils.isNotEmpty( values ) )
        {
            strings.addAll( Arrays.asList( values ) );
            strings.removeIf( value -> StringUtils.isBlank( StringUtils.stripToNull( value ) ) );
        }
        return ( new ArrayList<>( new LinkedHashSet<>( strings ) ) );
    }

    public boolean isInitialized()
    {
        return initialized;
    }

    public boolean isNotInitialized()
    {
        return !initialized;
    }

    public void close()
    {
        // close hibernate thread
        if ( isInitialized() )
        {
            try
            {
                hibernateUtil.getSessionFactory()
                             .close();
            }
            catch ( HibernateException exception )
            {
                // nothing to do
            }
        }
    }

    public boolean isApplicationsDatabaseEmpty()
    {
        return ( getApplicationRows( null ) == NO_ELEMENTS );
    }

    public boolean isServersDatabaseEmpty()
    {
        return ( getServerRows( null ) == NO_ELEMENTS );
    }

    public boolean isServerGroupsDatabaseEmpty()
    {
        return ( getServerGroupsRows( null ) == NO_ELEMENTS );
    }

    public boolean isUsersDatabaseEmpty()
    {
        return ( getUsersRows( null ) == NO_ELEMENTS );
    }

    public List<IconData> getAllIcons()
    {
        return getIcons( new Long[0] );
    }

    public List<IconData> getIcons( Long... iconIds )
    {
        List<IconData> iconDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<IconData> criteriaQuery = criteriaBuilder.createQuery( IconData.class );
            Root<IconData> icon = criteriaQuery.from( IconData.class );
            criteriaQuery = criteriaQuery.select( icon );
            List<Long> values = getValues( iconIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( icon.get( IconData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            iconDatas = session.createQuery( criteriaQuery )
                               .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return iconDatas;
    }

    public List<IconData> getIcons( String... iconNames )
    {
        List<IconData> iconDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<IconData> criteriaQuery = criteriaBuilder.createQuery( IconData.class );
            Root<IconData> icon = criteriaQuery.from( IconData.class );
            criteriaQuery = criteriaQuery.select( icon );
            List<String> values = getValues( iconNames );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( icon.get( IconData_.name ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            iconDatas = session.createQuery( criteriaQuery )
                               .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return iconDatas;
    }

    public IconData getIcon( long iconId )
    {
        List<IconData> iconDatas = getIcons( iconId );
        if ( CollectionUtils.size( iconDatas ) == ONE_ELEMENT )
        {
            return iconDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public IconData getIcon( String iconName )
    {
        if ( StringUtils.isNotBlank( iconName ) )
        {
            List<IconData> iconDatas = getIcons( iconName );
            if ( CollectionUtils.size( iconDatas ) == ONE_ELEMENT )
            {
                return iconDatas.get( FIRST_ELEMENT );
            }
        }
        return null;
    }

    public IconData getIcon( IconData iconData )
    {
        if ( Objects.nonNull( iconData ) )
        {
            if ( Objects.nonNull( iconData.getId() ) && iconData.getId() > NEW_ELEMENT )
            {
                return getIcon( iconData.getId() );
            }
            else if ( StringUtils.isNotBlank( StringUtils.stripToNull( iconData.getName() ) ) )
            {
                return getIcon( iconData.getName() );
            }
        }
        return null;
    }

    public boolean updateIcon( IconData iconData )
    {
        boolean iconUpdated = false;
        if ( Objects.nonNull( getIcon( iconData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( iconData );
                transaction.commit();
                iconUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return iconUpdated;
    }

    public boolean updateIcon( long id, String name, byte[] icon )
    {
        return updateIcon( new IconData( id, name, icon ) );
    }

    public long addIcon( IconData iconData )
    {
        long iconId = IconData.UNDEFINED_ICON;
        if ( Objects.nonNull( iconData ) && Objects.nonNull( iconData.getId() ) && Objects.isNull( getIcon( iconData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                iconData.setId( Objects.isNull( iconData.getId() ) || iconData.getId() <= IconData.NEW_ICON ? null : iconData.getId() );
                iconId = (Long) session.save( iconData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                iconId = IconData.UNDEFINED_ICON;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return iconId;
    }

    public long addIcon( String iconName, byte[] icon )
    {
        return addIcon( new IconData( IconData.NEW_ICON, iconName, icon ) );
    }

    public boolean removeIcon( long iconId )
    {
        boolean iconRemoved = false;
        IconData iconData = getIcon( iconId );
        if ( Objects.nonNull( iconData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.delete( iconData );
                transaction.commit();
                iconRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return iconRemoved;
    }

    public boolean removeIcon( IconData iconData )
    {
        if ( Objects.nonNull( iconData ) && Objects.nonNull( iconData.getId() ) )
        {
            return removeIcon( iconData.getId() );
        }
        return false;
    }

    public List<ServerData> getAllServers()
    {
        return getServers( new Long[0] );
    }

    public List<ServerData> getServers( Long... serverIds )
    {
        List<ServerData> serverDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ServerData> criteriaQuery = criteriaBuilder.createQuery( ServerData.class );
            Root<ServerData> server = criteriaQuery.from( ServerData.class );
            criteriaQuery = criteriaQuery.select( server );
            List<Long> values = getValues( serverIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( server.get( ServerData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            serverDatas = session.createQuery( criteriaQuery )
                                 .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serverDatas;
    }

    public List<ServerData> getServers( String... serverNames )
    {
        List<ServerData> serverDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ServerData> criteriaQuery = criteriaBuilder.createQuery( ServerData.class );
            Root<ServerData> server = criteriaQuery.from( ServerData.class );
            criteriaQuery = criteriaQuery.select( server );
            List<String> values = getValues( serverNames );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( server.get( ServerData_.name ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            serverDatas = session.createQuery( criteriaQuery )
                                 .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serverDatas;
    }

    public List<Long> getServerWithMaxUsersIds( Long... serverIds )
    {
        List<Long> servIds = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<Long[]> criteriaQuery = criteriaBuilder.createQuery( Long[].class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            Path<ServerData> serverPath = userSession.join( UserSessionData_.serverData );
            criteriaQuery = criteriaQuery.multiselect( serverPath.get( ServerData_.id ), serverPath.get( ServerData_.maxUsers ) );
            List<Long> values = getValues( serverIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( serverPath.get( ServerData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            criteriaQuery = criteriaQuery.groupBy( serverPath.get( ServerData_.id ), userSession.get( UserSessionData_.id ) );
            criteriaQuery = criteriaQuery.having( criteriaBuilder.ge( criteriaBuilder.countDistinct( serverPath.get( ServerData_.id ) ), serverPath.get( ServerData_.maxUsers ) ) );
            servIds = session.createQuery( criteriaQuery )
                             .getResultList()
                             .stream()
                             .map( data -> data[0] )
                             .collect( Collectors.toList() );
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return servIds;
    }

    public List<ServerData> getServers( int skipRows, int limit, Map<ServerColumns, String> filterMap, List<String> orders )
    {
        List<ServerData> serverDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ServerData> criteriaQuery = criteriaBuilder.createQuery( ServerData.class );
            Root<ServerData> server = criteriaQuery.from( ServerData.class );
            criteriaQuery = criteriaQuery.select( server );
            criteriaQuery = ServerFilter.addFilter( criteriaBuilder, criteriaQuery, server, filterMap );
            criteriaQuery = ServerOrder.addOrder( criteriaBuilder, criteriaQuery, server, orders );
            serverDatas = session.createQuery( criteriaQuery )
                                 .setFirstResult( skipRows )
                                 .setMaxResults( limit )
                                 .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serverDatas;
    }

    public Long getServerRows( Map<ServerColumns, String> filterMap )
    {
        Long serverRows = NO_ELEMENTS;
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
            Root<ServerData> server = criteriaQuery.from( ServerData.class );
            criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( server ) );
            criteriaQuery = ServerFilter.addFilter( criteriaBuilder, criteriaQuery, server, filterMap );
            serverRows = session.createQuery( criteriaQuery )
                                .getSingleResult();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serverRows;
    }

    public ServerData getServer( long serverId )
    {
        List<ServerData> serverDatas = getServers( serverId );
        if ( CollectionUtils.size( serverDatas ) == ONE_ELEMENT )
        {
            return serverDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public ServerData getServer( String serverName )
    {
        if ( StringUtils.isNotBlank( serverName ) )
        {
            List<ServerData> serverDatas = getServers( serverName );
            if ( CollectionUtils.size( serverDatas ) == ONE_ELEMENT )
            {
                return serverDatas.get( FIRST_ELEMENT );
            }
        }
        return null;
    }

    public ServerData getServer( ServerData serverData )
    {
        if ( Objects.nonNull( serverData ) )
        {
            if ( Objects.nonNull( serverData.getId() ) && serverData.getId() > NEW_ELEMENT )
            {
                return getServer( serverData.getId() );
            }
            else if ( StringUtils.isBlank( StringUtils.stripToNull( serverData.getName() ) ) )
            {
                return getServer( serverData.getName() );
            }
        }
        return null;
    }

    public boolean updateServer( ServerData serverData )
    {
        boolean serverUpdated = false;
        if ( Objects.nonNull( serverData ) && StringUtils.isNotBlank( serverData.getName() ) && StringUtils.isNotBlank( serverData.getAddress() ) && Objects.nonNull( serverData.getId() )
            && Objects.nonNull( serverData.getXpraMinPort() ) && Objects.nonNull( serverData.getXpraMaxPort() ) && Objects.nonNull( serverData.getMaxUsers() )
            && Objects.nonNull( serverData.getSshPort() ) && Objects.nonNull( getServer( serverData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( serverData );
                transaction.commit();
                serverUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return serverUpdated;
    }

    public boolean updateServer( int id, String name, String note, String address, int sshPort, int xpraMinPort, int xpraMaxPort, String xpraVersion, int maxUsers, boolean enabled )
    {
        return updateServer( new ServerData( id, name, note, address, sshPort, xpraMinPort, xpraMaxPort, xpraVersion, maxUsers, enabled ) );
    }

    public long addServer( ServerData serverData )
    {
        long serverId = ServerData.UNDEFINED_SERVER;
        if ( Objects.nonNull( serverData ) && StringUtils.isNotBlank( serverData.getName() ) && StringUtils.isNotBlank( serverData.getAddress() ) && Objects.nonNull( serverData.getXpraMinPort() )
            && Objects.nonNull( serverData.getXpraMaxPort() ) && Objects.nonNull( serverData.getMaxUsers() ) && Objects.nonNull( serverData.getSshPort() )
            && Objects.isNull( getServer( serverData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                serverData.setId( Objects.isNull( serverData.getId() ) || serverData.getId() <= ServerData.NEW_SERVER ? null : serverData.getId() );
                serverId = (Long) session.save( serverData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                serverId = ServerData.UNDEFINED_SERVER;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return serverId;
    }

    public long addServer( String name, String note, String address, int sshPort, int xpraMinPort, int xpraMaxPort, String xpraVersion, int maxUsers, boolean enabled )
    {
        return addServer( new ServerData( ServerData.NEW_SERVER, name, note, address, sshPort, xpraMinPort, xpraMaxPort, xpraVersion, maxUsers, enabled ) );
    }

    public boolean removeServer( long serverId )
    {
        boolean serverRemoved = false;
        ServerData serverData = getServer( serverId );
        if ( Objects.nonNull( serverData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaDelete<ServersGroupServerData> criteriaDelete = criteriaBuilder.createCriteriaDelete( ServersGroupServerData.class );
                Root<ServersGroupServerData> serversGroupServerData = criteriaDelete.from( ServersGroupServerData.class );
                criteriaDelete = criteriaDelete.where( criteriaBuilder.equal( serversGroupServerData.get( ServersGroupServerData_.serverId ), serverData.getId() ) );

                Transaction transaction = session.beginTransaction();
                session.createQuery( criteriaDelete )
                       .executeUpdate();
                session.delete( serverData );
                transaction.commit();
                serverRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return serverRemoved;
    }

    public boolean removeServer( ServerData serverData )
    {
        if ( Objects.nonNull( serverData ) && Objects.nonNull( serverData.getId() ) )
        {
            return removeServer( serverData.getId() );
        }
        return false;
    }

    public List<ServersGroupData> getAllServerGroups()
    {
        return getServerGroups( new Long[0] );
    }

    public List<ServersGroupData> getServerGroups( Long... serversGroupIds )
    {
        List<ServersGroupData> serversGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ServersGroupData> criteriaQuery = criteriaBuilder.createQuery( ServersGroupData.class );
            Root<ServersGroupData> serversGroup = criteriaQuery.from( ServersGroupData.class );
            criteriaQuery = criteriaQuery.select( serversGroup );
            List<Long> values = getValues( serversGroupIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( serversGroup.get( ServersGroupData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            serversGroupDatas = session.createQuery( criteriaQuery )
                                       .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serversGroupDatas;
    }

    public List<ServersGroupData> getServerGroups( String... serversGroupNames )
    {
        List<ServersGroupData> serversGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ServersGroupData> criteriaQuery = criteriaBuilder.createQuery( ServersGroupData.class );
            Root<ServersGroupData> serversGroup = criteriaQuery.from( ServersGroupData.class );
            criteriaQuery = criteriaQuery.select( serversGroup );
            List<String> values = getValues( serversGroupNames );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( serversGroup.get( ServersGroupData_.name ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            serversGroupDatas = session.createQuery( criteriaQuery )
                                       .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serversGroupDatas;
    }

    public List<ServersGroupData> getServerGroups( ServerData serverData )
    {
        List<ServersGroupData> serversGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ServersGroupData> criteriaQuery = criteriaBuilder.createQuery( ServersGroupData.class );
            Root<ServersGroupData> serversGroup = criteriaQuery.from( ServersGroupData.class );
            criteriaQuery = criteriaQuery.select( serversGroup );
            if ( Objects.nonNull( serverData ) )
            {
                if ( Objects.nonNull( serverData.getId() ) && serverData.getId() > NEW_ELEMENT )
                {
                    criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( serversGroup.join( ServersGroupData_.serverDatas )
                                                                                            .get( ServerData_.id ),
                                                                                serverData.getId() ) );
                }
                else if ( StringUtils.isNotBlank( StringUtils.stripToNull( serverData.getName() ) ) )
                {
                    criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( serversGroup.join( ServersGroupData_.serverDatas )
                                                                                            .get( ServerData_.name ),
                                                                                serverData.getName() ) );
                }
            }
            serversGroupDatas = session.createQuery( criteriaQuery )
                                       .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serversGroupDatas;
    }

    public List<ServersGroupData> getServerGroups( int skipRows, int limit, Map<ServersGroupColumns, String> filterMap, List<String> orders )
    {
        List<ServersGroupData> serversGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ServersGroupData> criteriaQuery = criteriaBuilder.createQuery( ServersGroupData.class );
            Root<ServersGroupData> serversGroup = criteriaQuery.from( ServersGroupData.class );
            criteriaQuery = criteriaQuery.select( serversGroup );
            criteriaQuery = ServersGroupFilter.addFilter( criteriaBuilder, criteriaQuery, serversGroup, filterMap );
            criteriaQuery = ServersGroupOrder.addOrder( criteriaBuilder, criteriaQuery, serversGroup, orders );
            criteriaQuery = criteriaQuery.groupBy( serversGroup );
            serversGroupDatas = session.createQuery( criteriaQuery )
                                       .setFirstResult( skipRows )
                                       .setMaxResults( limit )
                                       .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serversGroupDatas;
    }

    public Long getServerGroupsRows( Map<ServersGroupColumns, String> filterMap )
    {
        Long serversGroupRows = NO_ELEMENTS;
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
            Root<ServersGroupData> serversGroup = criteriaQuery.from( ServersGroupData.class );
            criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( serversGroup ) );
            criteriaQuery = ServersGroupFilter.addFilter( criteriaBuilder, criteriaQuery, serversGroup, filterMap );
            serversGroupRows = session.createQuery( criteriaQuery )
                                      .getSingleResult();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return serversGroupRows;
    }

    public ServersGroupData getServersGroup( long serversGroupId )
    {
        List<ServersGroupData> serversGroupDatas = getServerGroups( serversGroupId );
        if ( CollectionUtils.size( serversGroupDatas ) == ONE_ELEMENT )
        {
            return serversGroupDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public ServersGroupData getServersGroup( String serversGroupName )
    {
        if ( StringUtils.isNotBlank( serversGroupName ) )
        {
            List<ServersGroupData> serversGroupDatas = getServerGroups( serversGroupName );
            if ( CollectionUtils.size( serversGroupDatas ) == ONE_ELEMENT )
            {
                return serversGroupDatas.get( FIRST_ELEMENT );
            }
        }
        return null;
    }

    public ServersGroupData getServersGroup( ServersGroupData serversGroupData )
    {
        if ( Objects.nonNull( serversGroupData ) )
        {
            if ( Objects.nonNull( serversGroupData.getId() ) && serversGroupData.getId() > NEW_ELEMENT )
            {
                return getServersGroup( serversGroupData.getId() );
            }
            else if ( StringUtils.isNotBlank( StringUtils.stripToNull( serversGroupData.getName() ) ) )
            {
                return getServersGroup( serversGroupData.getName() );
            }
        }
        return null;
    }

    public boolean updateServersGroup( ServersGroupData serversGroupData )
    {
        boolean serversGroupUpdated = false;
        if ( Objects.nonNull( serversGroupData ) && StringUtils.isNotBlank( serversGroupData.getName() ) && Objects.nonNull( serversGroupData.getId() )
            && CollectionUtils.isNotEmpty( serversGroupData.getServerDatas() ) && Objects.nonNull( getServersGroup( serversGroupData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( serversGroupData );
                transaction.commit();
                serversGroupUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return serversGroupUpdated;
    }

    public boolean updateServersGroup( int id, String name, String note, boolean enabled, List<ServerData> serverDatas )
    {
        return updateServersGroup( new ServersGroupData( id, name, note, enabled, serverDatas ) );
    }

    public long addServersGroup( ServersGroupData serversGroupData )
    {
        long serversGroupId = ServersGroupData.UNDEFINED_SERVERS_GROUP;
        if ( Objects.nonNull( serversGroupData ) && StringUtils.isNotBlank( serversGroupData.getName() ) && CollectionUtils.isNotEmpty( serversGroupData.getServerDatas() )
            && Objects.isNull( getServersGroup( serversGroupData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                serversGroupData.setId( Objects.isNull( serversGroupData.getId() ) || serversGroupData.getId() <= ServersGroupData.NEW_SERVERS_GROUP ? null : serversGroupData.getId() );
                serversGroupId = (Long) session.save( serversGroupData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                serversGroupId = ServersGroupData.UNDEFINED_SERVERS_GROUP;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return serversGroupId;
    }

    public long addServersGroup( String name, String note, boolean enabled, List<ServerData> serverDatas )
    {
        return addServersGroup( new ServersGroupData( ServersGroupData.NEW_SERVERS_GROUP, name, note, enabled, serverDatas ) );
    }

    public boolean removeServersGroup( long serversGroupId )
    {
        boolean serversGroupRemoved = false;
        ServersGroupData serversGroupData = getServersGroup( serversGroupId );
        if ( Objects.nonNull( serversGroupData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaDelete<ApplicationServersGroupData> criteriaDelete = criteriaBuilder.createCriteriaDelete( ApplicationServersGroupData.class );
                Root<ApplicationServersGroupData> applicationServersGroup = criteriaDelete.from( ApplicationServersGroupData.class );
                criteriaDelete = criteriaDelete.where( criteriaBuilder.equal( applicationServersGroup.get( ApplicationServersGroupData_.serversGroupId ), serversGroupData.getId() ) );

                Transaction transaction = session.beginTransaction();
                session.createQuery( criteriaDelete )
                       .executeUpdate();
                session.delete( serversGroupData );
                transaction.commit();
                serversGroupRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return serversGroupRemoved;
    }

    public boolean removeServersGroup( ServersGroupData serversGroupData )
    {
        if ( Objects.nonNull( serversGroupData ) && Objects.nonNull( serversGroupData.getId() ) )
        {
            return removeServersGroup( serversGroupData.getId() );
        }
        return false;
    }

    public Long getServersGroupServerRows( ServerData serverData )
    {
        Long serversGroupServerRows = NO_ELEMENTS;
        if ( Objects.nonNull( serverData ) && Objects.nonNull( serverData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
                Root<ServersGroupServerData> serversGroupServer = criteriaQuery.from( ServersGroupServerData.class );
                criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( serversGroupServer ) );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( serversGroupServer.get( ServersGroupServerData_.serverId ), serverData.getId() ) );
                serversGroupServerRows = session.createQuery( criteriaQuery )
                                                .getSingleResult();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return serversGroupServerRows;
    }

    public List<ApplicationData> getAllApplications()
    {
        return getApplications( new Long[0] );
    }

    public List<ApplicationData> getApplications( Long... applicationIds )
    {
        List<ApplicationData> applicationDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ApplicationData> criteriaQuery = criteriaBuilder.createQuery( ApplicationData.class );
            Root<ApplicationData> application = criteriaQuery.from( ApplicationData.class );
            criteriaQuery = criteriaQuery.select( application );
            List<Long> values = getValues( applicationIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( application.get( ApplicationData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            applicationDatas = session.createQuery( criteriaQuery )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationDatas;
    }

    public List<ApplicationData> getApplications( String... applicationNames )
    {
        List<ApplicationData> applicationDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ApplicationData> criteriaQuery = criteriaBuilder.createQuery( ApplicationData.class );
            Root<ApplicationData> application = criteriaQuery.from( ApplicationData.class );
            criteriaQuery = criteriaQuery.select( application );
            List<String> values = getValues( applicationNames );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( application.get( ApplicationData_.name ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            applicationDatas = session.createQuery( criteriaQuery )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationDatas;
    }

    public List<ApplicationData> getApplications( int skipRows, int limit, Map<ApplicationColumns, String> filterMap, List<String> orders )
    {
        List<ApplicationData> applicationDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ApplicationData> criteriaQuery = criteriaBuilder.createQuery( ApplicationData.class );
            Root<ApplicationData> application = criteriaQuery.from( ApplicationData.class );
            criteriaQuery = criteriaQuery.select( application );
            criteriaQuery = ApplicationFilter.addFilter( criteriaBuilder, criteriaQuery, application, filterMap );
            criteriaQuery = ApplicationOrder.addOrder( criteriaBuilder, criteriaQuery, application, orders );
            criteriaQuery = criteriaQuery.groupBy( application );
            applicationDatas = session.createQuery( criteriaQuery )
                                      .setFirstResult( skipRows )
                                      .setMaxResults( limit )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationDatas;
    }

    public Long getApplicationRows( Map<ApplicationColumns, String> filterMap )
    {
        Long applicationRows = NO_ELEMENTS;
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
            Root<ApplicationData> application = criteriaQuery.from( ApplicationData.class );
            criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( application ) );
            criteriaQuery = ApplicationFilter.addFilter( criteriaBuilder, criteriaQuery, application, filterMap );
            applicationRows = session.createQuery( criteriaQuery )
                                     .getSingleResult();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationRows;
    }

    public ApplicationData getApplication( long applicationId )
    {
        List<ApplicationData> applicationDatas = new ArrayList<>( getApplications( applicationId ) );
        if ( CollectionUtils.size( applicationDatas ) == ONE_ELEMENT )
        {
            return applicationDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public ApplicationData getApplication( String applicationName )
    {
        List<ApplicationData> applicationDatas = new ArrayList<>( getApplications( applicationName ) );
        if ( CollectionUtils.size( applicationDatas ) == ONE_ELEMENT )
        {
            return applicationDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public ApplicationData getApplication( ApplicationData applicationData )
    {
        if ( Objects.nonNull( applicationData ) )
        {
            if ( Objects.nonNull( applicationData.getId() ) && applicationData.getId() > NEW_ELEMENT )
            {
                return getApplication( applicationData.getId() );
            }
            else if ( StringUtils.isNotBlank( StringUtils.stripToNull( applicationData.getName() ) ) )
            {
                return getApplication( applicationData.getName() );
            }
        }
        return null;
    }

    public boolean updateApplication( ApplicationData applicationData )
    {
        boolean applicationUpdated = false;
        if ( Objects.nonNull( applicationData ) && StringUtils.isNotBlank( applicationData.getName() ) && StringUtils.isNotBlank( applicationData.getCommand() )
            && Objects.nonNull( applicationData.getId() ) && CollectionUtils.isNotEmpty( applicationData.getServersGroupDatas() ) && Objects.nonNull( getApplication( applicationData ) ) )
        {
            IconData iconData = applicationData.getIconData();
            if ( iconData.getId() == IconData.NEW_ICON || Objects.isNull( getIcon( iconData.getName() ) ) )
            {
                IconData data = getIcon( iconData.getName() );
                if ( Objects.isNull( data ) )
                {
                    long id = addIcon( iconData );
                    data = getIcon( id );
                }
                applicationData.setIconData( data );
            }

            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( applicationData );
                transaction.commit();
                applicationUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationUpdated;
    }

    public boolean updateApplication( int id, String name, String command, String commandOptions, String note, boolean serverAutoSelect, int oneInstance, boolean sameInstance, boolean enabled,
                                      IconData iconData, List<ServersGroupData> serversGroupDatas )
    {
        return updateApplication( new ApplicationData( id, name, command, commandOptions, note, serverAutoSelect, oneInstance, sameInstance, enabled, iconData, serversGroupDatas ) );
    }

    public long addApplication( ApplicationData applicationData )
    {
        long applicationId = ApplicationData.UNDEFINED_APPLICATION;
        if ( Objects.nonNull( applicationData ) && StringUtils.isNotBlank( applicationData.getName() ) && StringUtils.isNotBlank( applicationData.getCommand() )
            && CollectionUtils.isNotEmpty( applicationData.getServersGroupDatas() ) && Objects.isNull( getApplication( applicationData ) ) )
        {
            String iconName = DEFAULT_ICON.create()
                                          .getElement()
                                          .getAttribute( ATTRIBUTE_ICON );
            IconData iconData = applicationData.getIconData();
            iconData = Objects.requireNonNullElse( iconData, new IconData( IconData.NEW_ICON, iconName, null ) );
            if ( Objects.isNull( iconData.getId() ) || iconData.getId() == IconData.NEW_ICON || Objects.isNull( getIcon( iconData.getName() ) ) )
            {
                IconData data = getIcon( iconData.getName() );
                if ( Objects.isNull( data ) )
                {
                    long id = addIcon( iconData );
                    data = getIcon( id );
                }
                applicationData.setIconData( data );
            }

            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                applicationData.setId( Objects.isNull( applicationData.getId() ) || applicationData.getId() <= ApplicationData.NEW_APPLICATION ? null : applicationData.getId() );
                applicationId = (Long) session.save( applicationData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                applicationId = ApplicationData.UNDEFINED_APPLICATION;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationId;
    }

    public long addApplication( String name, String command, String commandOptions, String note, boolean serverAutoSelect, int oneInstance, boolean sameInstance, boolean enabled, IconData iconData,
                                List<ServersGroupData> serversGroupDatas )
    {
        return addApplication( new ApplicationData( ApplicationData.NEW_APPLICATION, name, command, commandOptions, note, serverAutoSelect, oneInstance, sameInstance, enabled, iconData,
                                                    serversGroupDatas ) );
    }

    public boolean removeApplication( long applicationId )
    {
        boolean applicationRemoved = false;
        ApplicationData applicationData = getApplication( applicationId );
        if ( Objects.nonNull( applicationData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteria1Builder = session.getCriteriaBuilder();
                CriteriaDelete<ApplicationServersGroupData> criteria1Delete = criteria1Builder.createCriteriaDelete( ApplicationServersGroupData.class );
                Root<ApplicationServersGroupData> applicationServersGroup = criteria1Delete.from( ApplicationServersGroupData.class );
                criteria1Delete = criteria1Delete.where( criteria1Builder.equal( applicationServersGroup.get( ApplicationServersGroupData_.applicationId ), applicationData.getId() ) );

                CriteriaBuilder criteria2Builder = session.getCriteriaBuilder();
                CriteriaDelete<ApplicationsGroupApplicationData> criteria2Delete = criteria2Builder.createCriteriaDelete( ApplicationsGroupApplicationData.class );
                Root<ApplicationsGroupApplicationData> applicationsGroupApplication = criteria2Delete.from( ApplicationsGroupApplicationData.class );
                criteria2Delete = criteria2Delete.where( criteria2Builder.equal( applicationsGroupApplication.get( ApplicationsGroupApplicationData_.applicationId ), applicationData.getId() ) );

                CriteriaBuilder criteria3Builder = session.getCriteriaBuilder();
                CriteriaDelete<UserApplicationData> criteria3Delete = criteria3Builder.createCriteriaDelete( UserApplicationData.class );
                Root<UserApplicationData> userApplication = criteria3Delete.from( UserApplicationData.class );
                criteria3Delete = criteria3Delete.where( criteria3Builder.equal( userApplication.get( UserApplicationData_.applicationData ), applicationData ) );

                Transaction transaction = session.beginTransaction();
                session.createQuery( criteria1Delete )
                       .executeUpdate();
                session.createQuery( criteria2Delete )
                       .executeUpdate();
                session.createQuery( criteria3Delete )
                       .executeUpdate();
                session.delete( applicationData );
                transaction.commit();
                applicationRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationRemoved;
    }

    public boolean removeApplication( ApplicationData applicationData )
    {
        if ( Objects.nonNull( applicationData ) && Objects.nonNull( applicationData.getId() ) )
        {
            return removeApplication( applicationData.getId() );
        }
        return false;
    }

    public Long getApplicationServersGroupRows( ServersGroupData serversGroupData )
    {
        Long applicationServersGroupRows = NO_ELEMENTS;
        if ( Objects.nonNull( serversGroupData ) && Objects.nonNull( serversGroupData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
                Root<ApplicationServersGroupData> applicationServersGroup = criteriaQuery.from( ApplicationServersGroupData.class );
                criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( applicationServersGroup ) );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( applicationServersGroup.get( ApplicationServersGroupData_.serversGroupId ), serversGroupData.getId() ) );
                applicationServersGroupRows = session.createQuery( criteriaQuery )
                                                     .getSingleResult();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationServersGroupRows;
    }

    public List<ApplicationsGroupData> getAllApplicationGroups()
    {
        return getApplicationGroups( new Long[0] );
    }

    public List<ApplicationsGroupData> getApplicationGroups( Long... applicationsGroupIds )
    {
        List<ApplicationsGroupData> applicationsGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( ApplicationsGroupData.class );
            Root<ApplicationsGroupData> applicationsGroup = criteriaQuery.from( ApplicationsGroupData.class );
            criteriaQuery = criteriaQuery.select( applicationsGroup );
            List<Long> values = getValues( applicationsGroupIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( applicationsGroup.get( ApplicationsGroupData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            applicationsGroupDatas = session.createQuery( criteriaQuery )
                                            .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationsGroupDatas;
    }

    public List<ApplicationsGroupData> getApplicationGroups( ApplicationData applicationData )
    {
        List<ApplicationsGroupData> applicationsGroupDatas = new ArrayList<>();
        if ( Objects.nonNull( applicationData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<ApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( ApplicationsGroupData.class );
                Root<ApplicationsGroupData> applicationsGroup = criteriaQuery.from( ApplicationsGroupData.class );
                criteriaQuery = criteriaQuery.select( applicationsGroup );
                if ( Objects.nonNull( applicationData.getId() ) && applicationData.getId() > NEW_ELEMENT )
                {
                    criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( applicationsGroup.join( ApplicationsGroupData_.applicationDatas )
                                                                                                 .get( ApplicationData_.id ),
                                                                                applicationData.getId() ) );
                }
                else if ( StringUtils.isNotBlank( StringUtils.stripToNull( applicationData.getName() ) ) )
                {
                    criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( applicationsGroup.join( ApplicationsGroupData_.applicationDatas )
                                                                                                 .get( ApplicationData_.name ),
                                                                                applicationData.getName() ) );
                }
                criteriaQuery = criteriaQuery.groupBy( applicationsGroup );
                applicationsGroupDatas = session.createQuery( criteriaQuery )
                                                .getResultList();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationsGroupDatas;
    }

    public List<ApplicationsGroupData> getApplicationGroups( String... applicationsGroupNames )
    {
        List<ApplicationsGroupData> applicationsGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( ApplicationsGroupData.class );
            Root<ApplicationsGroupData> applicationsGroup = criteriaQuery.from( ApplicationsGroupData.class );
            criteriaQuery = criteriaQuery.select( applicationsGroup );
            List<String> values = getValues( applicationsGroupNames );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( applicationsGroup.get( ApplicationsGroupData_.name ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            applicationsGroupDatas = session.createQuery( criteriaQuery )
                                            .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationsGroupDatas;
    }

    public List<ApplicationsGroupData> getApplicationGroups( int skipRows, int limit, Map<ApplicationsGroupColumns, String> filterMap, List<String> orders )
    {
        List<ApplicationsGroupData> applicationsGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( ApplicationsGroupData.class );
            Root<ApplicationsGroupData> applicationsGroup = criteriaQuery.from( ApplicationsGroupData.class );
            criteriaQuery = criteriaQuery.select( applicationsGroup );
            criteriaQuery = ApplicationsGroupFilter.addFilter( criteriaBuilder, criteriaQuery, applicationsGroup, filterMap );
            criteriaQuery = ApplicationsGroupOrder.addOrder( criteriaBuilder, criteriaQuery, applicationsGroup, orders );
            criteriaQuery = criteriaQuery.groupBy( applicationsGroup );
            applicationsGroupDatas = session.createQuery( criteriaQuery )
                                            .setFirstResult( skipRows )
                                            .setMaxResults( limit )
                                            .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationsGroupDatas;
    }

    public Long getApplicationsGroupRows( Map<ApplicationsGroupColumns, String> filterMap )
    {
        Long applicationsGroupRows = NO_ELEMENTS;
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
            Root<ApplicationsGroupData> applicationsGroup = criteriaQuery.from( ApplicationsGroupData.class );
            criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( applicationsGroup ) );
            criteriaQuery = ApplicationsGroupFilter.addFilter( criteriaBuilder, criteriaQuery, applicationsGroup, filterMap );
            applicationsGroupRows = session.createQuery( criteriaQuery )
                                           .getSingleResult();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return applicationsGroupRows;
    }

    public ApplicationsGroupData getApplicationsGroup( long applicationsGroupId )
    {
        List<ApplicationsGroupData> applicationsGroupDatas = getApplicationGroups( applicationsGroupId );
        if ( CollectionUtils.size( applicationsGroupDatas ) == ONE_ELEMENT )
        {
            return applicationsGroupDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public ApplicationsGroupData getApplicationsGroup( String applicationsGroupName )
    {
        if ( StringUtils.isNotBlank( applicationsGroupName ) )
        {
            List<ApplicationsGroupData> applicationsGroupDatas = getApplicationGroups( applicationsGroupName );
            if ( CollectionUtils.size( applicationsGroupDatas ) == ONE_ELEMENT )
            {
                return applicationsGroupDatas.get( FIRST_ELEMENT );
            }
        }
        return null;
    }

    public ApplicationsGroupData getApplicationsGroup( ApplicationsGroupData applicationsGroupData )
    {
        if ( Objects.nonNull( applicationsGroupData ) )
        {
            if ( Objects.nonNull( applicationsGroupData.getId() ) && applicationsGroupData.getId() > NEW_ELEMENT )
            {
                return getApplicationsGroup( applicationsGroupData.getId() );
            }
            else if ( StringUtils.isNotBlank( StringUtils.stripToNull( applicationsGroupData.getName() ) ) )
            {
                return getApplicationsGroup( applicationsGroupData.getName() );
            }
        }
        return null;
    }

    public List<ApplicationsGroupData> getApplicationGroupsWithSameInstance( ApplicationData applicationData )
    {
        List<ApplicationsGroupData> applicationsGroupDatas = new ArrayList<>();
        if ( Objects.nonNull( applicationData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<ApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( ApplicationsGroupData.class );
                Root<ApplicationsGroupData> applicationsGroup = criteriaQuery.from( ApplicationsGroupData.class );
                Predicate applicationPredicate = criteriaBuilder.equal( applicationsGroup.join( ApplicationsGroupData_.applicationDatas )
                                                                                         .get( ApplicationData_.name ),
                                                                        applicationData.getName() );
                Predicate sameInstancePredicate = criteriaBuilder.equal( applicationsGroup.get( ApplicationsGroupData_.sameInstance ), Boolean.TRUE );
                criteriaQuery = criteriaQuery.select( applicationsGroup );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.and( applicationPredicate, sameInstancePredicate ) );
                criteriaQuery = criteriaQuery.groupBy( applicationsGroup );
                applicationsGroupDatas = session.createQuery( criteriaQuery )
                                                .getResultList();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationsGroupDatas;
    }

    public boolean isApplicationGroupsWithSameInstanceEmpty()
    {
        List<ApplicationsGroupData> applicationsGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<ApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( ApplicationsGroupData.class );
            Root<ApplicationsGroupData> applicationsGroup = criteriaQuery.from( ApplicationsGroupData.class );
            criteriaQuery = criteriaQuery.select( applicationsGroup );
            criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( applicationsGroup.get( ApplicationsGroupData_.sameInstance ), Boolean.TRUE ) );
            criteriaQuery = criteriaQuery.groupBy( applicationsGroup );
            applicationsGroupDatas = session.createQuery( criteriaQuery )
                                            .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return CollectionUtils.isEmpty( applicationsGroupDatas );
    }

    public boolean updateApplicationsGroup( ApplicationsGroupData applicationsGroupData )
    {
        boolean applicationsGroupUpdated = false;
        if ( Objects.nonNull( applicationsGroupData ) && StringUtils.isNotBlank( applicationsGroupData.getName() ) && Objects.nonNull( applicationsGroupData.getId() )
            && CollectionUtils.isNotEmpty( applicationsGroupData.getApplicationDatas() ) && Objects.nonNull( getApplicationsGroup( applicationsGroupData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( applicationsGroupData );
                transaction.commit();
                applicationsGroupUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationsGroupUpdated;
    }

    public boolean updateApplicationsGroup( long id, String name, String note, boolean sameInstance, boolean enabled, List<ApplicationData> applicationDatas )
    {
        return updateApplicationsGroup( new ApplicationsGroupData( id, name, note, sameInstance, enabled, applicationDatas ) );
    }

    public long addApplicationsGroup( ApplicationsGroupData applicationsGroupData )
    {
        long applicationsGroupId = ApplicationsGroupData.UNDEFINED_APPLICATIONS_GROUP;
        if ( Objects.nonNull( applicationsGroupData ) && StringUtils.isNotBlank( applicationsGroupData.getName() ) && CollectionUtils.isNotEmpty( applicationsGroupData.getApplicationDatas() )
            && Objects.isNull( getApplicationsGroup( applicationsGroupData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                applicationsGroupData.setId( Objects.isNull( applicationsGroupData.getId() ) || applicationsGroupData.getId() <= ApplicationsGroupData.NEW_APPLICATIONS_GROUP ? null
                                : applicationsGroupData.getId() );
                applicationsGroupId = (Long) session.save( applicationsGroupData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                applicationsGroupId = ApplicationsGroupData.UNDEFINED_APPLICATIONS_GROUP;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationsGroupId;
    }

    public long addApplicationsGroup( String name, String note, boolean sameInstance, boolean enabled, List<ApplicationData> applicationDatas )
    {
        return addApplicationsGroup( new ApplicationsGroupData( ApplicationsGroupData.NEW_APPLICATIONS_GROUP, name, note, sameInstance, enabled, applicationDatas ) );
    }

    public boolean removeApplicationsGroup( long applicationsGroupId )
    {
        boolean applicationsGroupRemoved = false;
        ApplicationsGroupData applicationsGroupData = getApplicationsGroup( applicationsGroupId );
        if ( Objects.nonNull( applicationsGroupData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteria1Builder = session.getCriteriaBuilder();
                CriteriaDelete<ApplicationsGroupApplicationData> criteria1Delete = criteria1Builder.createCriteriaDelete( ApplicationsGroupApplicationData.class );
                Root<ApplicationsGroupApplicationData> applicationsGroupApplication = criteria1Delete.from( ApplicationsGroupApplicationData.class );
                criteria1Delete =
                    criteria1Delete.where( criteria1Builder.equal( applicationsGroupApplication.get( ApplicationsGroupApplicationData_.applicationsGroupId ), applicationsGroupData.getId() ) );

                CriteriaBuilder criteria2Builder = session.getCriteriaBuilder();
                CriteriaDelete<UserApplicationsGroupData> criteria2Delete = criteria2Builder.createCriteriaDelete( UserApplicationsGroupData.class );
                Root<UserApplicationsGroupData> userApplicationsGroup = criteria2Delete.from( UserApplicationsGroupData.class );
                criteria2Delete = criteria2Delete.where( criteria2Builder.equal( userApplicationsGroup.get( UserApplicationsGroupData_.applicationsGroupData ), applicationsGroupData ) );

                Transaction transaction = session.beginTransaction();
                session.createQuery( criteria1Delete )
                       .executeUpdate();
                session.createQuery( criteria2Delete )
                       .executeUpdate();
                session.delete( applicationsGroupData );
                transaction.commit();
                applicationsGroupRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationsGroupRemoved;
    }

    public boolean removeApplicationsGroup( ApplicationsGroupData applicationsGroupData )
    {
        if ( Objects.nonNull( applicationsGroupData ) )
        {
            return removeApplicationsGroup( applicationsGroupData.getId() );
        }
        return false;
    }

    public Long getApplicationsGroupApplicationRows( ApplicationData applicationData )
    {
        Long applicationsGroupApplicationRows = NO_ELEMENTS;
        if ( Objects.nonNull( applicationData ) && Objects.nonNull( applicationData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
                Root<ApplicationsGroupApplicationData> applicationsGroupApplication = criteriaQuery.from( ApplicationsGroupApplicationData.class );
                criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( applicationsGroupApplication ) );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( applicationsGroupApplication.get( ApplicationsGroupApplicationData_.applicationId ), applicationData.getId() ) );
                applicationsGroupApplicationRows = session.createQuery( criteriaQuery )
                                                          .getSingleResult();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationsGroupApplicationRows;
    }

    public List<UserData> getAllUsers()
    {
        return getUsers( new Long[0] );
    }

    public List<UserData> getUsers( Long... userIds )
    {
        List<UserData> userDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserData> criteriaQuery = criteriaBuilder.createQuery( UserData.class );
            Root<UserData> user = criteriaQuery.from( UserData.class );
            criteriaQuery = criteriaQuery.select( user );
            List<Long> values = getValues( userIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( user.get( UserData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userDatas = session.createQuery( criteriaQuery )
                               .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userDatas;
    }

    public List<UserData> getUsers( String... userNames )
    {
        List<UserData> userDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserData> criteriaQuery = criteriaBuilder.createQuery( UserData.class );
            Root<UserData> user = criteriaQuery.from( UserData.class );
            criteriaQuery = criteriaQuery.select( user );
            List<String> values = getValues( userNames );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( user.get( UserData_.userName ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userDatas = session.createQuery( criteriaQuery )
                               .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userDatas;
    }

    public List<UserData> getUsers( int skipRows, int limit, Map<UserColumns, String> filterMap, List<String> orders )
    {
        List<UserData> userDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserData> criteriaQuery = criteriaBuilder.createQuery( UserData.class );
            Root<UserData> user = criteriaQuery.from( UserData.class );
            criteriaQuery = criteriaQuery.select( user );
            criteriaQuery = UserFilter.addFilter( criteriaBuilder, criteriaQuery, user, filterMap );
            criteriaQuery = UserOrder.addOrder( criteriaBuilder, criteriaQuery, user, orders );
            criteriaQuery = criteriaQuery.groupBy( user );
            userDatas = session.createQuery( criteriaQuery )
                               .setFirstResult( skipRows )
                               .setMaxResults( limit )
                               .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userDatas;
    }

    public Long getUsersRows( Map<UserColumns, String> filterMap )
    {
        Long userRows = NO_ELEMENTS;
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
            Root<UserData> user = criteriaQuery.from( UserData.class );
            criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( user ) );
            criteriaQuery = UserFilter.addFilter( criteriaBuilder, criteriaQuery, user, filterMap );
            userRows = session.createQuery( criteriaQuery )
                              .getSingleResult();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userRows;
    }

    public UserData getUser( long userId )
    {
        List<UserData> userDatas = getUsers( userId );
        if ( CollectionUtils.size( userDatas ) == ONE_ELEMENT )
        {
            return userDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public UserData getUser( String userName )
    {
        if ( StringUtils.isNotBlank( userName ) )
        {
            List<UserData> userDatas = getUsers( userName );
            if ( CollectionUtils.size( userDatas ) == ONE_ELEMENT )
            {
                return userDatas.get( FIRST_ELEMENT );
            }
        }
        return null;
    }

    public UserData getUser( UserData userData )
    {
        if ( Objects.nonNull( userData ) )
        {
            if ( Objects.nonNull( userData.getId() ) && userData.getId() > NEW_ELEMENT )
            {
                return getUser( userData.getId() );
            }
            else if ( StringUtils.isNotBlank( StringUtils.stripToNull( userData.getUserName() ) ) )
            {
                return getUser( userData.getUserName() );
            }
        }
        return null;
    }

    public boolean updateUser( UserData userData )
    {
        boolean userUpdated = false;
        if ( Objects.nonNull( userData ) && StringUtils.isNotBlank( userData.getUserName() ) && Objects.nonNull( userData.getId() ) && Objects.nonNull( userData.getMaxApplications() )
            && Objects.nonNull( getUser( userData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( userData );
                transaction.commit();
                userUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userUpdated;
    }

    public boolean updateUser( int id, String userName, String fullName, int maxApplications, int maxFileSize, boolean administrator, boolean screenshot, boolean upload, boolean download,
                               boolean fullScreen, boolean enabled )
    {
        return updateUser( new UserData( id, userName, fullName, maxApplications, maxFileSize, administrator, screenshot, upload, download, fullScreen, enabled ) );
    }

    public long addUser( UserData userData )
    {
        long userId = UserData.UNDEFINED_USER;
        if ( Objects.nonNull( userData ) && StringUtils.isNotBlank( userData.getUserName() ) && Objects.nonNull( userData.getMaxApplications() ) && Objects.isNull( getUser( userData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                userData.setId( Objects.isNull( userData.getId() ) || userData.getId() <= UserData.NEW_USER ? null : userData.getId() );
                userId = (Long) session.save( userData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                userId = UserData.UNDEFINED_USER;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userId;
    }

    public long addUser( String userName, String fullName, int maxApplications, int maxFileSize, boolean administrator, boolean screenshot, boolean upload, boolean download, boolean fullScreen,
                         boolean enabled )
    {
        return addUser( new UserData( UserData.NEW_USER, userName, fullName, maxApplications, maxFileSize, administrator, screenshot, upload, download, fullScreen, enabled ) );
    }

    public boolean removeUser( long userId )
    {
        boolean userRemoved = false;
        UserData userData = getUser( userId );
        if ( Objects.nonNull( userData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.delete( userData );
                transaction.commit();
                userRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userRemoved;
    }

    public boolean removeUser( UserData userData )
    {
        if ( Objects.nonNull( userData ) && Objects.nonNull( userData.getId() ) )
        {
            return removeUser( userData.getId() );
        }
        return false;
    }

    public List<UserApplicationData> getAllUserApplications()
    {
        return getUserApplications( new Long[0] );
    }

    public List<UserApplicationData> getUserApplications( Long... userApplicationIds )
    {
        List<UserApplicationData> userApplicationDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserApplicationData> criteriaQuery = criteriaBuilder.createQuery( UserApplicationData.class );
            Root<UserApplicationData> userApplication = criteriaQuery.from( UserApplicationData.class );
            criteriaQuery = criteriaQuery.select( userApplication );
            List<Long> values = getValues( userApplicationIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( userApplication.get( UserApplicationData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userApplicationDatas = session.createQuery( criteriaQuery )
                                          .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userApplicationDatas;
    }

    public List<UserApplicationData> getUserApplications( UserApplicationData userApplicationData )
    {
        List<UserApplicationData> userApplicationDatas = new ArrayList<>();
        if ( Objects.nonNull( userApplicationData ) && Objects.nonNull( userApplicationData.getId() ) )
        {
            userApplicationDatas = getUserApplications( userApplicationData.getId() );
        }
        return userApplicationDatas;
    }

    public List<UserApplicationData> getUserApplications( UserData userData )
    {
        List<UserApplicationData> userApplicationDatas = new ArrayList<>();
        if ( Objects.nonNull( userData ) && Objects.nonNull( userData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<UserApplicationData> criteriaQuery = criteriaBuilder.createQuery( UserApplicationData.class );
                Root<UserApplicationData> userApplication = criteriaQuery.from( UserApplicationData.class );
                criteriaQuery = criteriaQuery.select( userApplication );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( userApplication.get( UserApplicationData_.userData ), userData ) );
                userApplicationDatas = session.createQuery( criteriaQuery )
                                              .getResultList();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationDatas;
    }

    public UserApplicationData getUserApplication( long userApplicationId )
    {
        List<UserApplicationData> userApplicationDatas = getUserApplications( userApplicationId );
        if ( CollectionUtils.size( userApplicationDatas ) == ONE_ELEMENT )
        {
            return userApplicationDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public UserApplicationData getUserApplication( UserApplicationData userApplicationData )
    {
        List<UserApplicationData> userApplicationDatas = getUserApplications( userApplicationData );
        if ( CollectionUtils.size( userApplicationDatas ) == ONE_ELEMENT )
        {
            return userApplicationDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public Long getUserApplicationRows( ApplicationData applicationData )
    {
        Long applicationRows = NO_ELEMENTS;
        if ( Objects.nonNull( applicationData ) && Objects.nonNull( applicationData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
                Root<UserApplicationData> userApplication = criteriaQuery.from( UserApplicationData.class );
                criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( userApplication ) );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( userApplication.get( UserApplicationData_.applicationData ), applicationData ) );
                applicationRows = session.createQuery( criteriaQuery )
                                         .getSingleResult();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationRows;
    }

    public boolean updateUserApplication( UserApplicationData userApplicationData )
    {
        boolean userApplicationUpdated = false;
        if ( Objects.nonNull( userApplicationData ) && Objects.nonNull( userApplicationData.getId() ) && Objects.nonNull( userApplicationData.getApplicationData() )
            && Objects.nonNull( userApplicationData.getUserData() ) && Objects.nonNull( getUserApplication( userApplicationData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( userApplicationData );
                transaction.commit();
                userApplicationUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationUpdated;
    }

    public boolean updateUserApplication( long id, UserData userData, ApplicationData applicationData )
    {
        return updateUserApplication( new UserApplicationData( id, userData, applicationData ) );
    }

    public long addUserApplication( UserApplicationData userApplicationData )
    {
        long userApplicationId = UserApplicationData.UNDEFINED_USER_APPLICATION;
        if ( Objects.nonNull( userApplicationData ) && Objects.nonNull( userApplicationData.getId() ) && Objects.nonNull( userApplicationData.getApplicationData() )
            && Objects.nonNull( userApplicationData.getUserData() ) && Objects.isNull( getUserApplication( userApplicationData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                userApplicationData.setId( Objects.isNull( userApplicationData.getId() ) || userApplicationData.getId() <= UserApplicationData.NEW_USER_APPLICATION ? null
                                : userApplicationData.getId() );
                userApplicationId = (Long) session.save( userApplicationData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                userApplicationId = UserApplicationData.UNDEFINED_USER_APPLICATION;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationId;
    }

    public long addUserApplication( UserData userData, ApplicationData applicationData )
    {
        return addUserApplication( new UserApplicationData( UserApplicationData.NEW_USER_APPLICATION, userData, applicationData ) );
    }

    public boolean removeUserApplication( UserApplicationData userApplicationData )
    {
        boolean userApplicationRemoved = false;
        if ( Objects.nonNull( userApplicationData ) && Objects.nonNull( userApplicationData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.delete( userApplicationData );
                transaction.commit();
                userApplicationRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationRemoved;
    }

    public boolean removeUserApplication( long id, UserData userData, ApplicationData applicationData )
    {
        return removeUserApplication( new UserApplicationData( id, userData, applicationData ) );
    }

    public List<UserApplicationsGroupData> getAllUserApplicationGroups()
    {
        return getUserApplicationGroups( new Long[0] );
    }

    public List<UserApplicationsGroupData> getUserApplicationGroups( Long... userApplicationsGroupIds )
    {
        List<UserApplicationsGroupData> userApplicationsGroupDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( UserApplicationsGroupData.class );
            Root<UserApplicationsGroupData> userApplicationsGroup = criteriaQuery.from( UserApplicationsGroupData.class );
            criteriaQuery = criteriaQuery.select( userApplicationsGroup );
            List<Long> values = getValues( userApplicationsGroupIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( userApplicationsGroup.get( UserApplicationsGroupData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userApplicationsGroupDatas = session.createQuery( criteriaQuery )
                                                .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userApplicationsGroupDatas;
    }

    public List<UserApplicationsGroupData> getUserApplicationGroups( UserApplicationsGroupData userApplicationsGroupData )
    {
        List<UserApplicationsGroupData> userApplicationsGroupDatas = new ArrayList<>();
        if ( Objects.nonNull( userApplicationsGroupData ) && Objects.nonNull( userApplicationsGroupData.getId() ) )
        {
            userApplicationsGroupDatas = getUserApplicationGroups( userApplicationsGroupData.getId() );
        }
        return userApplicationsGroupDatas;
    }

    public List<UserApplicationsGroupData> getUserApplicationGroups( ApplicationsGroupData applicationsGroupData )
    {
        List<UserApplicationsGroupData> userApplicationsGroupDatas = new ArrayList<>();
        if ( Objects.nonNull( applicationsGroupData ) && Objects.nonNull( applicationsGroupData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<UserApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( UserApplicationsGroupData.class );
                Root<UserApplicationsGroupData> userApplicationsGroup = criteriaQuery.from( UserApplicationsGroupData.class );
                criteriaQuery = criteriaQuery.select( userApplicationsGroup );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( userApplicationsGroup.get( UserApplicationsGroupData_.applicationsGroupData ), applicationsGroupData ) );
                userApplicationsGroupDatas = session.createQuery( criteriaQuery )
                                                    .getResultList();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationsGroupDatas;
    }

    public List<UserApplicationsGroupData> getUserApplicationGroups( UserData userData )
    {
        List<UserApplicationsGroupData> userApplicationsGroupDatas = new ArrayList<>();
        if ( Objects.nonNull( userData ) && Objects.nonNull( userData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<UserApplicationsGroupData> criteriaQuery = criteriaBuilder.createQuery( UserApplicationsGroupData.class );
                Root<UserApplicationsGroupData> userApplicationsGroup = criteriaQuery.from( UserApplicationsGroupData.class );
                criteriaQuery = criteriaQuery.select( userApplicationsGroup );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( userApplicationsGroup.get( UserApplicationsGroupData_.userData ), userData ) );
                userApplicationsGroupDatas = session.createQuery( criteriaQuery )
                                                    .getResultList();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationsGroupDatas;
    }

    public UserApplicationsGroupData getUserApplicationsGroup( long userApplicationsGroupId )
    {
        List<UserApplicationsGroupData> userApplicationsGroupDatas = getUserApplicationGroups( userApplicationsGroupId );
        if ( CollectionUtils.size( userApplicationsGroupDatas ) == ONE_ELEMENT )
        {
            return userApplicationsGroupDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public UserApplicationsGroupData getUserApplicationsGroup( UserApplicationsGroupData userApplicationsGroupData )
    {
        List<UserApplicationsGroupData> userApplicationsGroupDatas = getUserApplicationGroups( userApplicationsGroupData );
        if ( CollectionUtils.size( userApplicationsGroupDatas ) == ONE_ELEMENT )
        {
            return userApplicationsGroupDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public Long getUserApplicationsGroupRows( ApplicationsGroupData applicationsGroupData )
    {
        Long applicationRows = NO_ELEMENTS;
        if ( Objects.nonNull( applicationsGroupData ) && Objects.nonNull( applicationsGroupData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<Long> criteriaQuery = criteriaBuilder.createQuery( Long.class );
                Root<UserApplicationsGroupData> userApplicationsGroup = criteriaQuery.from( UserApplicationsGroupData.class );
                criteriaQuery = criteriaQuery.multiselect( criteriaBuilder.countDistinct( userApplicationsGroup ) );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( userApplicationsGroup.get( UserApplicationsGroupData_.applicationsGroupData ), applicationsGroupData ) );
                applicationRows = session.createQuery( criteriaQuery )
                                         .getSingleResult();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return applicationRows;
    }

    public boolean updateUserApplicationsGroup( UserApplicationsGroupData userApplicationsGroupData )
    {
        boolean userApplicationsGroupUpdated = false;
        if ( Objects.nonNull( userApplicationsGroupData ) && Objects.nonNull( userApplicationsGroupData.getId() ) && Objects.nonNull( userApplicationsGroupData.getApplicationsGroupData() )
            && Objects.nonNull( userApplicationsGroupData.getUserData() ) && Objects.nonNull( getUserApplicationsGroup( userApplicationsGroupData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( userApplicationsGroupData );
                transaction.commit();
                userApplicationsGroupUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationsGroupUpdated;
    }

    public boolean updateUserApplicationsGroup( long id, UserData userData, ApplicationsGroupData applicationsGroupData )
    {
        return updateUserApplicationsGroup( new UserApplicationsGroupData( id, userData, applicationsGroupData ) );
    }

    public long addUserApplicationsGroup( UserApplicationsGroupData userApplicationsGroupData )
    {
        long userApplicationsGroupId = UserApplicationsGroupData.UNDEFINED_USER_APPLICATIONS_GROUP;
        if ( Objects.nonNull( userApplicationsGroupData ) && Objects.nonNull( userApplicationsGroupData.getId() ) && Objects.nonNull( userApplicationsGroupData.getApplicationsGroupData() )
            && Objects.nonNull( userApplicationsGroupData.getUserData() ) && Objects.isNull( getUserApplicationsGroup( userApplicationsGroupData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                userApplicationsGroupData.setId( Objects.isNull( userApplicationsGroupData.getId() ) || userApplicationsGroupData.getId() <= UserApplicationsGroupData.NEW_USER_APPLICATIONS_GROUP
                                ? null
                                : userApplicationsGroupData.getId() );
                userApplicationsGroupId = (Long) session.save( userApplicationsGroupData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                userApplicationsGroupId = UserApplicationsGroupData.UNDEFINED_USER_APPLICATIONS_GROUP;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationsGroupId;
    }

    public long addUserApplicationsGroup( UserData userData, ApplicationsGroupData applicationsGroupData )
    {
        return addUserApplicationsGroup( new UserApplicationsGroupData( UserApplicationsGroupData.NEW_USER_APPLICATIONS_GROUP, userData, applicationsGroupData ) );
    }

    public boolean removeUserApplicationsGroup( UserApplicationsGroupData userApplicationsGroupData )
    {
        boolean userApplicationsGroupRemoved = false;
        if ( Objects.nonNull( userApplicationsGroupData ) && Objects.nonNull( userApplicationsGroupData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.delete( userApplicationsGroupData );
                transaction.commit();
                userApplicationsGroupRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userApplicationsGroupRemoved;
    }

    public boolean removeUserApplicationsGroup( long id, UserData userData, ApplicationsGroupData applicationsGroupData )
    {
        return removeUserApplicationsGroup( new UserApplicationsGroupData( id, userData, applicationsGroupData ) );
    }

    public List<UserSessionData> getAllUsersSessions( Long... usersId )
    {
        List<UserSessionData> userSessionDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserSessionData> criteriaQuery = criteriaBuilder.createQuery( UserSessionData.class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            criteriaQuery = criteriaQuery.select( userSession );
            List<Long> values = getValues( usersId );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( userSession.get( UserSessionData_.userData )
                                                                        .get( UserData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userSessionDatas = session.createQuery( criteriaQuery )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userSessionDatas;
    }

    public List<UserSessionData> getAllUserSessions( long userId )
    {
        return getAllUsersSessions( userId );
    }

    public List<UserSessionData> getAllUserSessions( UserData userData )
    {
        if ( Objects.nonNull( userData ) && Objects.nonNull( userData.getId() ) )
        {
            return getAllUserSessions( userData.getId() );
        }
        return ( new ArrayList<>() );
    }

    public List<UserSessionData> getAllUserSessions( String... canvasIds )
    {
        List<UserSessionData> userSessionDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserSessionData> criteriaQuery = criteriaBuilder.createQuery( UserSessionData.class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            criteriaQuery = criteriaQuery.select( userSession );
            List<String> values = getValues( canvasIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( userSession.get( UserSessionData_.canvasId ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userSessionDatas = session.createQuery( criteriaQuery )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userSessionDatas;
    }

    public List<UserSessionData> getAllUserSessions( Map<UserSessionColumns, String> filterMap, String... canvasIds )
    {
        List<UserSessionData> userSessionDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserSessionData> criteriaQuery = criteriaBuilder.createQuery( UserSessionData.class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            criteriaQuery = criteriaQuery.select( userSession );
            criteriaQuery = UserSessionFilter.addFilter( criteriaBuilder, criteriaQuery, userSession, filterMap );
            List<String> values = getValues( canvasIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<String> inRestriction = criteriaBuilder.in( userSession.get( UserSessionData_.canvasId ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userSessionDatas = session.createQuery( criteriaQuery )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userSessionDatas;
    }

    public List<String> getAllUserSessionCanvasIds( int skipRows, int limit, Map<UserSessionColumns, String> filterMap )
    {
        List<String> canvasIds = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<String> criteriaQuery = criteriaBuilder.createQuery( String.class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            criteriaQuery = criteriaQuery.multiselect( userSession.get( UserSessionData_.canvasId ) )
                                         .distinct( true );
            criteriaQuery = UserSessionFilter.addFilter( criteriaBuilder, criteriaQuery, userSession, filterMap );
            canvasIds = session.createQuery( criteriaQuery )
                               .setFirstResult( skipRows )
                               .setMaxResults( limit )
                               .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return canvasIds;
    }

    public Integer getAllUserSessionCanvasIdsRows( Map<UserSessionColumns, String> filterMap )
    {
        Integer usersSessionsRows = (int) NO_ELEMENTS;
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<String> criteriaQuery = criteriaBuilder.createQuery( String.class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            criteriaQuery = criteriaQuery.multiselect( userSession.get( UserSessionData_.canvasId ) )
                                         .distinct( true );
            criteriaQuery = UserSessionFilter.addFilter( criteriaBuilder, criteriaQuery, userSession, filterMap );
            usersSessionsRows = session.createQuery( criteriaQuery )
                                       .getResultList()
                                       .size();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return usersSessionsRows;
    }

    public UserSessionData getUserSession( long userSessionId )
    {
        UserSessionData userSessionData = null;
        List<UserSessionData> userSessionDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserSessionData> criteriaQuery = criteriaBuilder.createQuery( UserSessionData.class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            criteriaQuery = criteriaQuery.select( userSession );
            In<Long> inRestriction = criteriaBuilder.in( userSession.get( UserSessionData_.id ) );
            criteriaQuery = criteriaQuery.where( inRestriction.value( userSessionId ) );
            userSessionDatas = session.createQuery( criteriaQuery )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }

        if ( CollectionUtils.size( userSessionDatas ) == ONE_ELEMENT )
        {
            userSessionData = userSessionDatas.get( FIRST_ELEMENT );
        }

        return userSessionData;
    }

    public UserSessionData getUserSession( UserSessionData userSessionData )
    {
        if ( Objects.nonNull( userSessionData ) && Objects.nonNull( userSessionData.getId() ) )
        {
            return getUserSession( userSessionData.getId() );
        }
        return null;
    }

    public UserSessionData getUserSession( long serverId, int port )
    {
        UserSessionData userSessionData = null;
        List<UserSessionData> userSessionDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserSessionData> criteriaQuery = criteriaBuilder.createQuery( UserSessionData.class );
            Root<UserSessionData> userSession = criteriaQuery.from( UserSessionData.class );
            Predicate serverPredicate = criteriaBuilder.equal( userSession.get( UserSessionData_.serverData )
                                                                          .get( ServerData_.id ),
                                                               serverId );
            Predicate portPredicate = criteriaBuilder.equal( userSession.get( UserSessionData_.port ), port );
            criteriaQuery = criteriaQuery.select( userSession );
            criteriaQuery = criteriaQuery.where( criteriaBuilder.and( serverPredicate, portPredicate ) );
            userSessionDatas = session.createQuery( criteriaQuery )
                                      .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }

        if ( CollectionUtils.size( userSessionDatas ) == ONE_ELEMENT )
        {
            userSessionData = userSessionDatas.get( FIRST_ELEMENT );
        }
        return userSessionData;
    }

    public boolean updateUserSession( UserSessionData userSessionData )
    {
        boolean userSessionUpdated = false;
        if ( Objects.nonNull( userSessionData ) && Objects.nonNull( userSessionData.getId() ) && Objects.nonNull( userSessionData.getUserData() ) && Objects.nonNull( userSessionData.getServerData() )
            && Objects.nonNull( userSessionData.getPort() ) && Objects.nonNull( userSessionData.getDesktopWidth() ) && Objects.nonNull( userSessionData.getDesktopHeight() )
            && StringUtils.isNotBlank( userSessionData.getCanvasId() ) && Objects.nonNull( getUserSession( userSessionData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.update( userSessionData );
                transaction.commit();
                userSessionUpdated = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionUpdated;
    }

    public long addDemoUserSession( long userId, long serverId, long applicationId, int port )
    {
        long userSessionId = UserSessionData.UNDEFINED_USER_SESSION;
        UserSessionData userSessionData = getUserSession( serverId, port );
        if ( Objects.isNull( userSessionData ) )
        {
            UserData userData = getUser( userId );
            ServerData serverData = getServer( serverId );
            ApplicationData applicationData = getApplication( applicationId );
            userSessionData = new UserSessionData( UserSessionData.NEW_USER_SESSION, userData, serverData, port, applicationData, UserSessionData.NEW_DATA, UserSessionData.NEW_CANVAS_ID,
                                                   UserSessionData.DESKTOP_WIDTH, UserSessionData.DESKTOP_HEIGHT, UserSessionData.NEW_WINDOW, false, null );
            userSessionData.setWindowId( null );
            userSessionId = addUserSession( userSessionData );
        }
        return userSessionId;
    }

    public long addUserSession( UserSessionData userSessionData )
    {
        long userSessionId = UserSessionData.UNDEFINED_USER_SESSION;
        if ( Objects.nonNull( userSessionData ) && Objects.nonNull( userSessionData.getUserData() ) && Objects.nonNull( userSessionData.getServerData() )
            && Objects.nonNull( userSessionData.getPort() ) && Objects.nonNull( userSessionData.getDesktopWidth() ) && Objects.nonNull( userSessionData.getDesktopHeight() )
            && StringUtils.isNotBlank( userSessionData.getCanvasId() ) && Objects.isNull( getUserSession( userSessionData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                userSessionData.setId( Objects.isNull( userSessionData.getId() ) || userSessionData.getId() <= UserSessionData.NEW_USER_SESSION ? null : userSessionData.getId() );
                userSessionData.setWindowId( UserSessionData.UNDEFINED_WINDOW == userSessionData.getWindowId() ? null : userSessionData.getWindowId() );
                userSessionData.setUpdated( new Date() );
                userSessionId = (Long) session.save( userSessionData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                userSessionId = UserSessionData.UNDEFINED_USER_SESSION;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionId;
    }

    public long addUserSession( UserData userData, ServerData serverData, int port, ApplicationData applicationData, String status, String canvasId, int desktopWidth, int desktopHeight, int windowId,
                                boolean locked )
    {
        if ( StringUtils.isNotBlank( canvasId ) )
        {
            return addUserSession( new UserSessionData( UserSessionData.NEW_USER_SESSION, userData, serverData, port, applicationData, status, canvasId, desktopWidth, desktopHeight, windowId, locked,
                                                        null ) );
        }
        return UserSessionData.UNDEFINED_USER_SESSION;
    }

    public boolean removeUserSession( long userSessionId )
    {
        boolean userSessionRemoved = false;
        UserSessionData userSessionData = getUserSession( userSessionId );
        if ( Objects.nonNull( userSessionData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.delete( userSessionData );
                transaction.commit();
                userSessionRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionRemoved;
    }

    public boolean removeUserSession( UserSessionData userSessionData )
    {
        if ( Objects.nonNull( userSessionData ) && Objects.nonNull( userSessionData.getId() ) )
        {
            return removeUserSession( userSessionData.getId() );
        }
        return false;
    }

    public List<UserSessionApplicationData> getAllUserSessionApplications()
    {
        return getUserSessionApplications( new Long[0] );
    }

    public List<UserSessionApplicationData> getUserSessionApplications( Long... userSessionApplicationIds )
    {
        List<UserSessionApplicationData> userSessionApplicationDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserSessionApplicationData> criteriaQuery = criteriaBuilder.createQuery( UserSessionApplicationData.class );
            Root<UserSessionApplicationData> userSessionApplication = criteriaQuery.from( UserSessionApplicationData.class );
            criteriaQuery = criteriaQuery.select( userSessionApplication );
            List<Long> values = getValues( userSessionApplicationIds );
            if ( CollectionUtils.isNotEmpty( values ) )
            {
                In<Long> inRestriction = criteriaBuilder.in( userSessionApplication.get( UserSessionApplicationData_.id ) );
                values.forEach( value -> inRestriction.value( value ) );
                criteriaQuery = criteriaQuery.where( inRestriction );
            }
            userSessionApplicationDatas = session.createQuery( criteriaQuery )
                                                 .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }
        return userSessionApplicationDatas;
    }

    public List<UserSessionApplicationData> getUserSessionApplications( UserSessionData userSessionData )
    {
        List<UserSessionApplicationData> userSessionApplicationDatas = new ArrayList<>();
        if ( Objects.nonNull( userSessionData ) && Objects.nonNull( userSessionData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<UserSessionApplicationData> criteriaQuery = criteriaBuilder.createQuery( UserSessionApplicationData.class );
                Root<UserSessionApplicationData> userSessionApplication = criteriaQuery.from( UserSessionApplicationData.class );
                criteriaQuery = criteriaQuery.select( userSessionApplication );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( userSessionApplication.get( UserSessionApplicationData_.userSessionId ), userSessionData.getId() ) );
                userSessionApplicationDatas = session.createQuery( criteriaQuery )
                                                     .getResultList();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionApplicationDatas;
    }

    public List<UserSessionApplicationData> getUserSessionApplications( UserData userData )
    {
        List<UserSessionApplicationData> userSessionApplicationDatas = new ArrayList<>();
        if ( Objects.nonNull( userData ) && Objects.nonNull( userData.getId() ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaQuery<UserSessionApplicationData> criteriaQuery = criteriaBuilder.createQuery( UserSessionApplicationData.class );
                Root<UserSessionApplicationData> userSessionApplication = criteriaQuery.from( UserSessionApplicationData.class );
                criteriaQuery = criteriaQuery.select( userSessionApplication );
                criteriaQuery = criteriaQuery.where( criteriaBuilder.equal( userSessionApplication.get( UserSessionApplicationData_.userId ), userData.getId() ) );
                userSessionApplicationDatas = session.createQuery( criteriaQuery )
                                                     .getResultList();
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionApplicationDatas;
    }

    public UserSessionApplicationData getUserSessionApplication( long userSessionApplicationId )
    {
        List<UserSessionApplicationData> userSessionApplicationDatas = getUserSessionApplications( userSessionApplicationId );
        if ( CollectionUtils.size( userSessionApplicationDatas ) == ONE_ELEMENT )
        {
            return userSessionApplicationDatas.get( FIRST_ELEMENT );
        }
        return null;
    }

    public UserSessionApplicationData getUserSessionApplication( long userSessionId, int windowId )
    {
        UserSessionApplicationData userSessionApplicationData = null;
        List<UserSessionApplicationData> userSessionApplicationDatas = new ArrayList<>();
        try (Session session = hibernateUtil.getSessionFactory()
                                            .openSession())
        {
            CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
            CriteriaQuery<UserSessionApplicationData> criteriaQuery = criteriaBuilder.createQuery( UserSessionApplicationData.class );
            Root<UserSessionApplicationData> userSessionApplication = criteriaQuery.from( UserSessionApplicationData.class );
            Predicate userSessionPredicate = criteriaBuilder.equal( userSessionApplication.get( UserSessionApplicationData_.userSessionId ), userSessionId );
            Predicate windowIdPredicate = criteriaBuilder.equal( userSessionApplication.get( UserSessionApplicationData_.windowId ), windowId );
            criteriaQuery = criteriaQuery.select( userSessionApplication )
                                         .where( criteriaBuilder.and( userSessionPredicate, windowIdPredicate ) );
            userSessionApplicationDatas = session.createQuery( criteriaQuery )
                                                 .getResultList();
        }
        catch ( Exception exception )
        {
            LOGGER.warn( exception.getMessage() );
        }

        if ( CollectionUtils.size( userSessionApplicationDatas ) == ONE_ELEMENT )
        {
            userSessionApplicationData = userSessionApplicationDatas.get( FIRST_ELEMENT );
        }
        return userSessionApplicationData;
    }

    public UserSessionApplicationData getUserSessionApplication( UserSessionApplicationData userSessionApplicationData )
    {
        if ( Objects.nonNull( userSessionApplicationData ) && Objects.nonNull( userSessionApplicationData.getId() ) )
        {
            return getUserSessionApplication( userSessionApplicationData.getId() );
        }
        return null;
    }

    public long addUserSessionApplication( UserSessionApplicationData userSessionApplicationData )
    {
        long userSessionApplicationId = UserSessionApplicationData.UNDEFINED_USER_SESSION_APPLICATION;
        if ( Objects.nonNull( userSessionApplicationData ) && Objects.nonNull( userSessionApplicationData.getUserSessionId() ) && Objects.nonNull( userSessionApplicationData.getApplicationId() )
            && Objects.nonNull( userSessionApplicationData.getWindowId() ) && Objects.nonNull( userSessionApplicationData.getUserId() )
            && Objects.isNull( getUserSessionApplication( userSessionApplicationData ) ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                userSessionApplicationData.setId( userSessionApplicationData.getId() <= UserSessionApplicationData.NEW_USER_SESSION_APPLICATION ? null : userSessionApplicationData.getId() );
                userSessionApplicationId = (Long) session.save( userSessionApplicationData );
                transaction.commit();
            }
            catch ( Exception exception )
            {
                userSessionApplicationId = UserSessionApplicationData.UNDEFINED_USER_SESSION_APPLICATION;
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionApplicationId;
    }

    public long addUserSessionApplication( long userSessionId, long applicationId, int windowId, long userId )
    {
        return addUserSessionApplication( new UserSessionApplicationData( UserSessionApplicationData.NEW_USER_SESSION_APPLICATION, userSessionId, applicationId, windowId, userId ) );
    }

    public boolean removeUserSessionApplication( long userSessionApplicationId )
    {
        boolean userSessionApplicationRemoved = false;
        UserSessionApplicationData userSessionApplicationData = getUserSessionApplication( userSessionApplicationId );
        if ( Objects.nonNull( userSessionApplicationData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.delete( userSessionApplicationData );
                transaction.commit();
                userSessionApplicationRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionApplicationRemoved;
    }

    public boolean removeUserSessionApplication( UserSessionApplicationData userSessionApplicationData )
    {
        if ( Objects.nonNull( userSessionApplicationData ) && Objects.nonNull( userSessionApplicationData.getId() ) )
        {
            return removeUserSessionApplication( userSessionApplicationData.getId() );
        }
        return false;
    }

    public boolean removeUserSessionApplication( UserSessionData userSessionData )
    {
        boolean userSessionApplicationRemoved = false;
        List<UserSessionApplicationData> userSessionApplicationDatas = getUserSessionApplications( userSessionData );
        if ( CollectionUtils.isNotEmpty( userSessionApplicationDatas ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                userSessionApplicationDatas.forEach( userSessionApplicationData -> session.delete( userSessionApplicationData ) );
                transaction.commit();
                userSessionApplicationRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionApplicationRemoved;
    }

    public boolean removeUserSessionApplication( long userSessionId, int windowId )
    {
        boolean userSessionApplicationRemoved = false;
        UserSessionApplicationData userSessionApplicationData = getUserSessionApplication( userSessionId, windowId );
        if ( Objects.nonNull( userSessionApplicationData ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                Transaction transaction = session.beginTransaction();
                session.delete( userSessionApplicationData );
                transaction.commit();
                userSessionApplicationRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionApplicationRemoved;
    }

    public boolean removeUserSessionUnactiveApplications( long userId )
    {
        boolean userSessionApplicationRemoved = false;
        if ( Objects.nonNull( userId ) )
        {
            try (Session session = hibernateUtil.getSessionFactory()
                                                .openSession())
            {
                List<UserSessionData> userSessionDatas = getAllUsersSessions( userId );
                CriteriaBuilder criteriaBuilder = session.getCriteriaBuilder();
                CriteriaDelete<UserSessionApplicationData> criteriaDelete = criteriaBuilder.createCriteriaDelete( UserSessionApplicationData.class );
                Root<UserSessionApplicationData> userSessionApplication = criteriaDelete.from( UserSessionApplicationData.class );
                if ( CollectionUtils.isEmpty( userSessionDatas ) )
                {
                    criteriaDelete = criteriaDelete.where( criteriaBuilder.equal( userSessionApplication.get( UserSessionApplicationData_.userId ), userId ) );
                }
                else
                {
                    In<Long> inRestriction = criteriaBuilder.in( userSessionApplication.get( UserSessionApplicationData_.userSessionId ) );
                    userSessionDatas.forEach( userSessionData -> inRestriction.value( userSessionData.getId() ) );
                    criteriaDelete = criteriaDelete.where( inRestriction.not() );
                }
                Transaction transaction = session.beginTransaction();
                session.createQuery( criteriaDelete )
                       .executeUpdate();
                transaction.commit();
                userSessionApplicationRemoved = true;
            }
            catch ( Exception exception )
            {
                LOGGER.warn( exception.getMessage() );
            }
        }
        return userSessionApplicationRemoved;
    }

}
