/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Objects;
import java.util.concurrent.ScheduledFuture;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;

import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpException;

public abstract class FileUploadRunner
    implements Runnable
{

    private static final String FILE_SEPARATOR = File.separator;

    private String srcUploadedFileName = null;

    private String dstUploadedFileName = null;

    private ScheduledFuture<?> scheduledFuture = null;

    private UserSessionData userSessionData = null;

    private VirtualUtil virtualUtil = null;

    public FileUploadRunner( VirtualUtil virtualUtil, UserSessionData userSessionData )
    {
        this.virtualUtil = virtualUtil;
        this.userSessionData = userSessionData;
    }

    public abstract void startUploadingFile();

    public abstract void stopUploadingFile( String fileName );

    @Override
    public void run()
    {
        // start uploading file
        startUploadingFile();

        ChannelSftp sftpChannel = null;
        Session session = null;
        if ( Objects.nonNull( virtualUtil ) && StringUtils.isNotBlank( srcUploadedFileName ) && StringUtils.isNotBlank( dstUploadedFileName ) )
        {
            // stop the scheduler
            scheduledFuture.cancel( false );
            scheduledFuture = null;

            // send the file from the client to the server
            try
            {
                // start SSH session
                session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                 userSessionData.getServerData()
                                                                .getAddress(),
                                                 userSessionData.getServerData()
                                                                .getSshPort() );
                session.connect();

                sftpChannel = (ChannelSftp) session.openChannel( Command.SSH.SFTP );
                sftpChannel.connect();
                sftpChannel.put( srcUploadedFileName, dstUploadedFileName );

                // delete file (srcUploadedFileName) from temporary folder
                Files.deleteIfExists( Paths.get( srcUploadedFileName ) );

                sftpChannel.disconnect();
                session.disconnect();
            }
            catch ( IllegalThreadStateException | IOException | JSchException | SftpException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( sftpChannel ) && sftpChannel.isConnected() )
                {
                    sftpChannel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }
        }

        // stop uploading file
        String[] path = dstUploadedFileName.split( FILE_SEPARATOR );
        stopUploadingFile( path[path.length - 1] );
    }

    public void setScheduledFuture( ScheduledFuture<?> scheduledFuture )
    {
        this.scheduledFuture = scheduledFuture;
    }

    public void setUploadedFileName( String srcUploadedFileName, String dstUploadedFileName )
    {
        this.srcUploadedFileName = srcUploadedFileName;
        this.dstUploadedFileName = dstUploadedFileName;
    }

}
