/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.applicationgroups;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationsGroupColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.button.ButtonVariant;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.value.ValueChangeMode;

public class ApplicationGroupsUI
    extends VerticalLayout
{

    private static final long serialVersionUID = -5388407159083971045L;

    private static final int FILTER_VALUE_CHANGE_TIMEOUT = 500;

    private static final int MAX_ELEMENTS_IN_ROW = 4;

    private static final int VISIBLE_ROWS = 50;

    private static final long NO_ELEMENTS = 0;

    private static final String CLASS_GRID_GROUPS = "grid-groups";

    private static final String CLASS_GRID_GROUPS_BUTTON = "grid-groups-button";

    private static final String CLASS_GRID_GROUPS_ICON = "grid-groups-icon";

    private static final String CLASS_GRID_GROUPS_LABEL = "grid-groups-label";

    private static final String CLASS_SETTINGS = "settings";

    private static final String NAME_WITH_NOTE = "%s (%s)";

    private static final String STYLE_COLOR = "color";

    private static final String STYLE_COLOR_RED_VALUE = "red";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_VALUE = "1px";

    private static final String STYLE_MARGIN_BOTTOM = "margin-bottom";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_TOP = "margin-top";

    private static final String STYLE_MARGIN_0PX_VALUE = "0px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_100PX_VALUE = "100px";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private transient List<ApplicationsGroupData> applicationsGroupDatas = null;

    private transient Map<ApplicationsGroupColumns, String> filterMap = null;

    private Database database = null;

    private DataProvider<ApplicationsGroupData, Void> dataProvider = null;

    private Grid<ApplicationsGroupData> applicationsGroupGrid = null;

    private Grid.Column<ApplicationsGroupData> editDeleteColumn = null;

    private VirtualUtil virtualUtil = null;

    private ApplicationGroupsUI()
    {
        setMargin( false );
        setSpacing( true );
        setSizeFull();
        addClassName( CLASS_SETTINGS );
        filterMap = new EnumMap<>( ApplicationsGroupColumns.class );
    }

    public ApplicationGroupsUI( Database database, VirtualUtil virtualUtil )
    {
        this();
        this.database = database;
        this.virtualUtil = virtualUtil;

        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( SortDirection.DESCENDING.equals( order.getDirection() ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            return new ArrayList<>( this.database.getApplicationGroups( offset, limit, filterMap, orders ) ).stream();
        }, query -> database.getApplicationsGroupRows( filterMap )
                            .intValue() );

        applicationsGroupDatas = new ArrayList<>( database.getAllApplicationGroups() );
        applicationsGroupGrid = new Grid<>();
        applicationsGroupGrid.setDataProvider( dataProvider );
        applicationsGroupGrid.setPageSize( VISIBLE_ROWS );
        applicationsGroupGrid.getStyle()
                             .set( STYLE_DISPLAY, STYLE_DISPLAY_VALUE )
                             .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                             .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE );
        applicationsGroupGrid.addItemDoubleClickListener( event -> {
            applicationsGroupGrid.select( event.getItem() );
            editApplicationsGroupData( event.getItem() ).open();
        } );
        add( applicationsGroupGrid );

        // name
        Grid.Column<ApplicationsGroupData> nameColumn = applicationsGroupGrid.addColumn( ApplicationsGroupData::getName )
                                                                             .setFlexGrow( 1 )
                                                                             .setResizable( true )
                                                                             .setSortable( true )
                                                                             .setSortProperty( ApplicationsGroupColumns.NAME.getName() )
                                                                             .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                                             .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.NAME.getValueForTranslation() ) );
        // note
        Grid.Column<ApplicationsGroupData> noteColumn = applicationsGroupGrid.addColumn( ApplicationsGroupData::getNote )
                                                                             .setFlexGrow( 3 )
                                                                             .setResizable( true )
                                                                             .setSortable( true )
                                                                             .setSortProperty( ApplicationsGroupColumns.NOTE.getName() )
                                                                             .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.NOTE.getValueForTranslation() ) );
        // application
        Grid.Column<ApplicationsGroupData> applicationColumn = applicationsGroupGrid.addColumn( new ComponentRenderer<>( Div::new, ( applicationsDiv, applicationsGroupData ) -> {
            List<ApplicationData> applicationDatas = new ArrayList<>( applicationsGroupData.getApplicationDatas() ).stream()
                                                                                                                   .sorted( Comparator.comparing( ApplicationData::getName ) )
                                                                                                                   .collect( Collectors.toList() );
            IntStream.range( 0, Math.min( MAX_ELEMENTS_IN_ROW, applicationDatas.size() ) )
                     .filter( index -> Objects.nonNull( applicationDatas.get( index ) ) )
                     .forEach( index -> {
                         Div applicationDiv = new Div();
                         ApplicationData applicationData = applicationDatas.get( index );
                         if ( index < ( MAX_ELEMENTS_IN_ROW - 1 ) || applicationDatas.size() <= MAX_ELEMENTS_IN_ROW )
                         {
                             Label label = new Label( StringUtils.isBlank( applicationData.getNote() ) ? applicationData.getName()
                                             : String.format( NAME_WITH_NOTE, applicationData.getName(), applicationData.getNote() ) );
                             label.addClassName( CLASS_GRID_GROUPS_LABEL );
                             Icon icon = applicationData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                             icon.addClassName( CLASS_GRID_GROUPS_ICON );

                             applicationDiv.addClassName( CLASS_GRID_GROUPS );
                             applicationDiv.add( icon );
                             applicationDiv.add( label );
                         }
                         else
                         {
                             Button applicationsButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_SHOW ), event -> {
                                 applicationsGroupGrid.select( applicationsGroupData );
                                 editApplicationsGroupData( applicationsGroupData ).open();
                             } );
                             applicationsButton.setWidthFull();
                             applicationsButton.addClassName( CLASS_GRID_GROUPS_BUTTON );
                             applicationsButton.addThemeVariants( ButtonVariant.LUMO_CONTRAST );
                             applicationDiv.addClassName( CLASS_GRID_GROUPS );
                             applicationDiv.add( applicationsButton );
                             applicationDiv.getStyle()
                                           .set( STYLE_MARGIN_BOTTOM, STYLE_MARGIN_0PX_VALUE )
                                           .set( STYLE_MARGIN_TOP, STYLE_MARGIN_0PX_VALUE );
                         }
                         applicationsDiv.add( applicationDiv );
                     } );
        } ) )
                                                                                    .setFlexGrow( 3 )
                                                                                    .setResizable( true )
                                                                                    .setSortable( false )
                                                                                    .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.APPLICATION.getValueForTranslation() ) );
        // same instance
        Grid.Column<ApplicationsGroupData> sameInstanceColumn = applicationsGroupGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationsGroupData ) -> {
            Icon icon = applicationsGroupData.isSameInstance() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                                     .setFlexGrow( 0 )
                                                                                     .setResizable( true )
                                                                                     .setSortable( true )
                                                                                     .setSortProperty( ApplicationsGroupColumns.SAME_INSTANCE.getName() )
                                                                                     .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                                                     .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.SAME_INSTANCE.getValueForTranslation() ) );
        // enabled
        Grid.Column<ApplicationsGroupData> enabledColumn = applicationsGroupGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationsGroupData ) -> {
            Icon icon = applicationsGroupData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                                .setFlexGrow( 0 )
                                                                                .setResizable( true )
                                                                                .setSortable( true )
                                                                                .setSortProperty( ApplicationsGroupColumns.ENABLED.getName() )
                                                                                .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                                                .setHeader( virtualUtil.getTranslation( ApplicationsGroupColumns.ENABLED.getValueForTranslation() ) );
        // edit/delete data
        editDeleteColumn = applicationsGroupGrid.addColumn( new ComponentRenderer<>( applicationsGroupData -> {
            // delete
            Button deleteButton = new Button( new Icon( VaadinIcon.TRASH ), event -> {
                boolean noElements = database.getUserApplicationsGroupRows( applicationsGroupData ) == NO_ELEMENTS;
                String message = noElements ? virtualUtil.getTranslation( Translation.APPLICATIONS_GROUP_REMOVE, applicationsGroupData.getName() )
                                : virtualUtil.getTranslation( Translation.APPLICATIONS_GROUP_REMOVE_USER, applicationsGroupData.getName() );
                ApplicationsGroupRemoveDialog dialog = new ApplicationsGroupRemoveDialog( message, virtualUtil )
                {
                    private static final long serialVersionUID = -1288407159083971045L;

                    @Override
                    public void savingData()
                    {
                        removeApplicationsGroup( applicationsGroupData );
                    }
                };
                dialog.open();
                applicationsGroupGrid.select( applicationsGroupData );
            } );
            deleteButton.getStyle()
                        .set( STYLE_COLOR, STYLE_COLOR_RED_VALUE );

            // edit
            Button editButton = new Button( new Icon( VaadinIcon.EDIT ), event -> {
                applicationsGroupGrid.select( applicationsGroupData );
                editApplicationsGroupData( applicationsGroupData ).open();
            } );

            if ( virtualUtil.isTooltipsVisible() )
            {
                TooltipRunner tooltipRunner1 = new TooltipRunner( deleteButton, virtualUtil.getTranslation( Translation.BUTTON_DELETE ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner1 );
                TooltipRunner tooltipRunner2 = new TooltipRunner( editButton, virtualUtil.getTranslation( Translation.BUTTON_EDIT ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner2 );
            }

            HorizontalLayout layout = new HorizontalLayout( editButton, deleteButton );
            layout.setMargin( false );
            layout.setPadding( false );
            layout.setSpacing( false );
            layout.setSizeUndefined();
            return layout;
        } ) )
                                                .setFlexGrow( 0 )
                                                .setWidth( STYLE_WIDTH_100PX_VALUE )
                                                .setResizable( false );
        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationsGroupDatas ) );

        // filter: name
        HeaderRow filterRow = applicationsGroupGrid.appendHeaderRow();
        TextField nameFilter = new TextField();
        nameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameFilter.setSizeFull();
        nameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        nameFilter.addValueChangeListener( event -> {
            if ( nameFilter.isEmpty() )
            {
                filterMap.remove( ApplicationsGroupColumns.NAME );
            }
            else
            {
                filterMap.put( ApplicationsGroupColumns.NAME, nameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( nameColumn )
                 .setComponent( nameFilter );

        // filter: note
        TextField noteFilter = new TextField();
        noteFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteFilter.setSizeFull();
        noteFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        noteFilter.addValueChangeListener( event -> {
            if ( noteFilter.isEmpty() )
            {
                filterMap.remove( ApplicationsGroupColumns.NOTE );
            }
            else
            {
                filterMap.put( ApplicationsGroupColumns.NOTE, noteFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( noteColumn )
                 .setComponent( noteFilter );

        // filter: applications
        TextField applicationFilter = new TextField();
        applicationFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        applicationFilter.setSizeFull();
        applicationFilter.setValueChangeMode( ValueChangeMode.LAZY );
        applicationFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        applicationFilter.addValueChangeListener( event -> {
            if ( applicationFilter.isEmpty() )
            {
                filterMap.remove( ApplicationsGroupColumns.APPLICATION );
            }
            else
            {
                filterMap.put( ApplicationsGroupColumns.APPLICATION, applicationFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( applicationColumn )
                 .setComponent( applicationFilter );

        // filter: same instance
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> sameInstanceFilter = new ComboBox<>();
        sameInstanceFilter.setAllowCustomValue( false );
        sameInstanceFilter.setClassName( STYLE_GRID_COMBOBOX );
        sameInstanceFilter.setItems( filterAll, filterYes, filterNo );
        sameInstanceFilter.setValue( filterAll );
        sameInstanceFilter.setSizeFull();
        sameInstanceFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( sameInstanceFilter.getValue(), filterYes ) )
            {
                filterMap.put( ApplicationsGroupColumns.SAME_INSTANCE, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( sameInstanceFilter.getValue(), filterNo ) )
            {
                filterMap.put( ApplicationsGroupColumns.SAME_INSTANCE, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ApplicationsGroupColumns.SAME_INSTANCE );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( sameInstanceColumn )
                 .setComponent( sameInstanceFilter );

        // filter: enabled
        ComboBox<String> enabledFilter = new ComboBox<>();
        enabledFilter.setAllowCustomValue( false );
        enabledFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledFilter.setItems( filterAll, filterYes, filterNo );
        enabledFilter.setValue( filterAll );
        enabledFilter.setSizeFull();
        enabledFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledFilter.getValue(), filterYes ) )
            {
                filterMap.put( ApplicationsGroupColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledFilter.getValue(), filterNo ) )
            {
                filterMap.put( ApplicationsGroupColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ApplicationsGroupColumns.ENABLED );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( enabledColumn )
                 .setComponent( enabledFilter );
    }

    private ApplicationsGroupEditDialog editApplicationsGroupData( ApplicationsGroupData applicationsGroupData )
    {
        return ( new ApplicationsGroupEditDialog( applicationsGroupData.getId(), applicationsGroupData, database, virtualUtil )
        {
            private static final long serialVersionUID = 5388407159083971045L;

            @Override
            public void savingData( ApplicationsGroupData data )
            {
                if ( Objects.nonNull( data ) )
                {
                    boolean closeDialog = true;
                    if ( Objects.nonNull( database.getApplicationsGroup( data.getId() ) ) )
                    {
                        closeDialog = updateApplicationsGroup( data );
                    }
                    else
                    {
                        closeDialog = addApplicationsGroup( data );
                    }

                    if ( closeDialog )
                    {
                        close();
                    }
                }
                else
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
            }
        } );
    }

    public boolean addApplicationsGroup( ApplicationsGroupData applicationsGroupData )
    {
        if ( Objects.nonNull( applicationsGroupData ) && Objects.isNull( database.getApplicationsGroup( applicationsGroupData.getId() ) ) )
        {
            applicationsGroupData.setId( database.addApplicationsGroup( applicationsGroupData ) );
            if ( !Objects.equals( applicationsGroupData.getId(), ApplicationsGroupData.UNDEFINED_APPLICATIONS_GROUP ) )
            {
                applicationsGroupDatas.add( applicationsGroupData );
                dataProvider.refreshAll();
                applicationsGroupGrid.select( applicationsGroupData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationsGroupDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATIONS_GROUP_ERROR_ADD ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean addApplicationsGroup( long id, String name, String note, boolean sameInstance, boolean enabled, List<ApplicationData> applicationDatas )
    {
        return addApplicationsGroup( new ApplicationsGroupData( id, name, note, sameInstance, enabled, applicationDatas ) );
    }

    public boolean removeApplicationsGroup( ApplicationsGroupData applicationsGroupData )
    {
        if ( Objects.nonNull( applicationsGroupData ) )
        {
            if ( database.removeApplicationsGroup( applicationsGroupData ) )
            {
                applicationsGroupGrid.deselectAll();
                applicationsGroupDatas.remove( applicationsGroupData );
                dataProvider.refreshAll();
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationsGroupDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATIONS_GROUP_ERROR_REMOVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean removeApplicationsGroup( int id )
    {
        return removeApplicationsGroup( database.getApplicationsGroup( id ) );
    }

    public boolean updateApplicationsGroup( ApplicationsGroupData applicationsGroupData )
    {
        if ( Objects.nonNull( applicationsGroupData ) )
        {
            if ( database.updateApplicationsGroup( applicationsGroupData ) )
            {
                applicationsGroupDatas.removeIf( data -> Objects.equals( data.getId(), applicationsGroupData.getId() ) );
                applicationsGroupDatas.add( applicationsGroupData );
                dataProvider.refreshAll();
                applicationsGroupGrid.select( applicationsGroupData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationsGroupDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE, applicationsGroupData.getName() ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean updateApplicationsGroup( long id, String name, String note, boolean sameInstance, boolean enabled, List<ApplicationData> applicationDatas )
    {
        return updateApplicationsGroup( new ApplicationsGroupData( id, name, note, sameInstance, enabled, applicationDatas ) );
    }

    public void addNewApplicationsGroup()
    {
        ApplicationsGroupEditDialog dialog = new ApplicationsGroupEditDialog( ApplicationsGroupData.NEW_APPLICATIONS_GROUP, new ApplicationsGroupData(), database, virtualUtil )
        {
            private static final long serialVersionUID = -1460198394796534539L;

            @Override
            public void savingData( ApplicationsGroupData applicationsGroupData )
            {
                if ( Objects.nonNull( applicationsGroupData ) )
                {
                    applicationsGroupData.setId( database.addApplicationsGroup( applicationsGroupData ) );
                    if ( !Objects.equals( applicationsGroupData.getId(), ApplicationsGroupData.UNDEFINED_APPLICATIONS_GROUP ) )
                    {
                        applicationsGroupDatas.add( applicationsGroupData );
                        dataProvider.refreshAll();
                        applicationsGroupGrid.select( applicationsGroupData );
                        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationsGroupDatas ) );
                        close();
                    }
                    else
                    {
                        new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATIONS_GROUP_ERROR_EXISTS, applicationsGroupData.getName() ),
                                            virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                    }
                }
                else
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
            }
        };
        dialog.open();
    }

}
