/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.applications;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.icons.BrandsIcon;
import org.dussan.vaadin.dvirtual.icons.EntypoIcon;
import org.dussan.vaadin.dvirtual.icons.FeatherIcon;
import org.dussan.vaadin.dvirtual.icons.IcomoonIcon;
import org.dussan.vaadin.dvirtual.icons.IconicIcon;
import org.dussan.vaadin.dvirtual.icons.JqueryuiIcon;
import org.dussan.vaadin.dvirtual.icons.TypIcon;
import org.dussan.vaadin.dvirtual.icons.WpzoomIcon;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogEditUtil;

import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.formlayout.FormLayout;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep;
import com.vaadin.flow.component.formlayout.FormLayout.ResponsiveStep.LabelsPosition;
import com.vaadin.flow.component.html.Span;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.listbox.ListBox;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.upload.Upload;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;

public abstract class ApplicationIconsDialog
    extends DialogEditUtil
{

    private static final long serialVersionUID = -8720878152766374267L;

    private static final int FORM_COLUMNS = 1;

    private static final int MAX_FILE_SIZE = 1000000;

    private static final int MAX_FILES = 1;

    private static final String ATTRIBUTE_CLASS_BUTTON_VALUE = "primary";

    private static final String ATTRIBUTE_ICON = "icon";

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String DIALOG_BUTTONS_CLASS = "edit-dialog-buttons";

    private static final String DIALOG_CLASS = "edit-dialog";

    private static final String FILE_TYPE_GIF = "image/gif";

    private static final String FILE_TYPE_JPEG = "image/jpeg";

    private static final String FILE_TYPE_JPG = "image/jpg";

    private static final String FILE_TYPE_PNG = "image/png";

    private static final String FORM_MINIMAL_WIDTH = "0";

    private static final String ICON_BORDER_CLASS = "icon-border";

    private static final String ICON_SELECTED_CLASS = "icon-selected";

    private static final String ICON_NAME_EMPTY = StringUtils.EMPTY;

    private static final String ICON_NAME_SEPARATOR = ":";

    private static final String ICONS_LAYOUT_CLASS = "icons-layout";

    private static final String STYLE_BORDER = "border";

    private static final String STYLE_BORDER_VALUE = "1px solid var(--lumo-body-text-color)";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_BLOCK_VALUE = "block";

    private static final String STYLE_DISPLAY_NONE_VALUE = "none";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_VALUE = "0px";

    private static final String STYLE_WIDTH_100PERCENT_VALUE = "100%";

    private static final String TEMPORARY_FOLDER = System.getProperty( "java.io.tmpdir" );

    private static final String VAADIN_ICON_COLLECTION_NAME = "vaadin";

    private static final String VAADIN_ICON_DESCRIPTION = "Vaadin";

    private String iconName = null;

    private Button importSelectFileButton = null;

    private Button selectButton = null;

    private Icon selectedIcon = null;

    private Upload upload = null;

    private VirtualUtil virtualUtil = null;

    public ApplicationIconsDialog( String iconName, VirtualUtil virtualUtil )
    {
        this.iconName = iconName;
        this.virtualUtil = virtualUtil;

        // select/close
        selectButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_ICON ) );
        selectButton.setEnabled( false );
        selectButton.addClassName( ATTRIBUTE_CLASS_BUTTON_VALUE );
        selectButton.addClickListener( event -> close() );
        selectButton.getElement()
                    .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        selectButton.getStyle()
                    .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE );
        Button closeButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_CLOSE ), event -> close() );
        closeButton.getElement()
                   .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        Button uploadButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_UPLOAD_ICON ), event -> {
            addImportDataComponent();
            importSelectFileButton.clickInClient();
        } );
        uploadButton.getElement()
                    .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );

        // buttons layout
        HorizontalLayout buttonsLayout = new HorizontalLayout( selectButton, uploadButton, closeButton );
        buttonsLayout.setJustifyContentMode( JustifyContentMode.CENTER );
        buttonsLayout.setMargin( true );
        buttonsLayout.setPadding( true );
        buttonsLayout.setSpacing( true );
        buttonsLayout.addClassName( DIALOG_BUTTONS_CLASS );

        // icons layout
        HorizontalLayout iconsLayout = new HorizontalLayout();
        iconsLayout.setMargin( false );
        iconsLayout.setPadding( false );
        iconsLayout.setSpacing( false );
        iconsLayout.setSizeUndefined();
        iconsLayout.setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        iconsLayout.getStyle()
                   .set( STYLE_DISPLAY, STYLE_DISPLAY_BLOCK_VALUE );

        VerticalLayout verticalLayout = new VerticalLayout( iconsLayout );
        verticalLayout.addClassName( ICONS_LAYOUT_CLASS );

        // icons
        String brandsIcons = virtualUtil.getTranslation( Translation.ICONS, BrandsIcon.ICON_DESCRIPTION );
        String entypoIcons = virtualUtil.getTranslation( Translation.ICONS, EntypoIcon.ICON_DESCRIPTION );
        String featherIcons = virtualUtil.getTranslation( Translation.ICONS, FeatherIcon.ICON_DESCRIPTION );
        String icomoonIcons = virtualUtil.getTranslation( Translation.ICONS, IcomoonIcon.ICON_DESCRIPTION );
        String iconicIcons = virtualUtil.getTranslation( Translation.ICONS, IconicIcon.ICON_DESCRIPTION );
        String jqueryuiIcons = virtualUtil.getTranslation( Translation.ICONS, JqueryuiIcon.ICON_DESCRIPTION );
        String typiconIcons = virtualUtil.getTranslation( Translation.ICONS, TypIcon.ICON_DESCRIPTION );
        String vaadinIcons = virtualUtil.getTranslation( Translation.ICONS, VAADIN_ICON_DESCRIPTION );
        String wpzoomIcons = virtualUtil.getTranslation( Translation.ICONS, WpzoomIcon.ICON_DESCRIPTION );
        List<Span> iconSpans = new ArrayList<>();
        ListBox<String> iconsListBox = new ListBox<>();
        iconsListBox.setItems( brandsIcons, entypoIcons, featherIcons, icomoonIcons, iconicIcons, jqueryuiIcons, typiconIcons, vaadinIcons, wpzoomIcons );
        iconsListBox.getElement()
                    .getStyle()
                    .set( STYLE_BORDER, STYLE_BORDER_VALUE );
        iconsListBox.addValueChangeListener( event -> {
            String value = event.getValue();
            iconSpans.clear();
            if ( StringUtils.equals( value, brandsIcons ) )
            {
                for ( BrandsIcon ic : BrandsIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, entypoIcons ) )
            {
                for ( EntypoIcon ic : EntypoIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, featherIcons ) )
            {
                for ( FeatherIcon ic : FeatherIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, icomoonIcons ) )
            {
                for ( IcomoonIcon ic : IcomoonIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, iconicIcons ) )
            {
                for ( IconicIcon ic : IconicIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, jqueryuiIcons ) )
            {
                for ( JqueryuiIcon ic : JqueryuiIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, typiconIcons ) )
            {
                for ( TypIcon ic : TypIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, wpzoomIcons ) )
            {
                for ( WpzoomIcon ic : WpzoomIcon.values() )
                {
                    Icon icon = ic.create();
                    icon.setId( ic.getIconName() );
                    setIconSelected( icon, getSelectedIconName().equals( ic.getIconName() ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }
            else if ( StringUtils.equals( value, vaadinIcons ) )
            {
                for ( VaadinIcon ic : VaadinIcon.values() )
                {
                    Icon icon = ic.create();
                    String name = icon.getElement()
                                      .getAttribute( ATTRIBUTE_ICON );
                    icon.setId( name );
                    setIconSelected( icon, getSelectedIconName().equals( name ) );
                    Span iconSpan = new Span( icon );
                    iconSpans.add( iconSpan );
                    iconSpan.addClickListener( evet -> setSelectedIcon( icon ) );
                }
            }

            // add icons
            iconsLayout.removeAll();
            iconsLayout.add( iconSpans.toArray( Span[]::new ) );
        } );

        // icon list box value
        switch ( getIconCollectionName( iconName ) )
        {
            case BrandsIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( brandsIcons );
                break;
            case EntypoIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( entypoIcons );
                break;
            case FeatherIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( featherIcons );
                break;
            case IcomoonIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( icomoonIcons );
                break;
            case IconicIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( iconicIcons );
                break;
            case JqueryuiIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( jqueryuiIcons );
                break;
            case TypIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( typiconIcons );
                break;
            case WpzoomIcon.ICON_COLLECTION_NAME:
                iconsListBox.setValue( wpzoomIcons );
                break;
            case VAADIN_ICON_COLLECTION_NAME:
            default:
                iconsListBox.setValue( vaadinIcons );
                break;
        }

        HorizontalLayout layout = new HorizontalLayout( iconsListBox, verticalLayout );
        layout.setMargin( false );
        layout.setPadding( false );
        layout.setSpacing( true );
        layout.expand( iconsLayout );

        // form layout
        setFormLayout( new FormLayout( layout, buttonsLayout ) );
        getFormLayout().setClassName( DIALOG_CLASS );
        getFormLayout().setResponsiveSteps( new ResponsiveStep( FORM_MINIMAL_WIDTH, FORM_COLUMNS, LabelsPosition.TOP ) );
        getFormLayout().setWidth( STYLE_WIDTH_100PERCENT_VALUE );
        add( getFormLayout() );
    }

    public abstract void importingData( String folder, String fileName );

    public abstract void importingDataError( String errorMessage );

    private void addImportDataComponent()
    {
        // only use for select the file for the import
        importSelectFileButton = new Button( virtualUtil.getTranslation( Translation.BUTTON_IMPORT ) );
        upload = new Upload( ( fileName, mimeType ) -> {
            File tmpFile = new File( TEMPORARY_FOLDER, fileName );
            try
            {
                return new FileOutputStream( tmpFile );
            }
            catch ( FileNotFoundException exception )
            {
                return null;
            }
        } );
        upload.setDropAllowed( false );
        upload.setAcceptedFileTypes( FILE_TYPE_GIF, FILE_TYPE_JPEG, FILE_TYPE_JPG, FILE_TYPE_PNG );
        upload.setMaxFiles( MAX_FILES );
        upload.setMaxFileSize( MAX_FILE_SIZE );
        upload.setUploadButton( importSelectFileButton );
        upload.setSizeUndefined();
        upload.getStyle()
              .set( STYLE_DISPLAY, STYLE_DISPLAY_NONE_VALUE );
        upload.addFailedListener( event -> importingDataError( event.getReason()
                                                                    .getMessage() ) );
        upload.addFileRejectedListener( event -> importingDataError( event.getErrorMessage() ) );
        upload.addSucceededListener( event -> importingData( TEMPORARY_FOLDER, event.getFileName() ) );
        add( upload );
    }

    private String getIconCollectionName( String iconName )
    {
        return iconName.split( ICON_NAME_SEPARATOR )[0];
    }

    private void setIconSelected( Icon icon, boolean selected )
    {
        icon.addClassName( selected ? ICON_SELECTED_CLASS : ICON_BORDER_CLASS );
        icon.removeClassName( !selected ? ICON_SELECTED_CLASS : ICON_BORDER_CLASS );
        selectedIcon = selected ? icon : selectedIcon;
    }

    public String getSelectedIconName()
    {
        if ( Objects.isNull( selectedIcon ) )
        {
            return iconName;
        }
        return selectedIcon.getId()
                           .orElse( ICON_NAME_EMPTY );
    }

    public Icon getSelectedIcon()
    {
        return selectedIcon;
    }

    public void setSelectedIcon( Icon icon )
    {
        if ( Objects.nonNull( icon ) )
        {
            if ( Objects.nonNull( selectedIcon ) )
            {
                setIconSelected( selectedIcon, false );
            }
            setIconSelected( icon, true );
            selectButton.setEnabled( true );
            selectedIcon = icon;
        }
    }

    public Button getSelectButton()
    {
        return selectButton;
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        if ( Objects.nonNull( upload ) )
        {
            remove( upload );
            importSelectFileButton = null;
            upload = null;
        }
    }

}
