/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.applications;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.columns.ApplicationColumns;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.combobox.ComboBox;
import com.vaadin.flow.component.grid.Grid;
import com.vaadin.flow.component.grid.HeaderRow;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Image;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.IronIcon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.data.provider.DataProvider;
import com.vaadin.flow.data.provider.SortDirection;
import com.vaadin.flow.data.provider.SortOrder;
import com.vaadin.flow.data.renderer.ComponentRenderer;
import com.vaadin.flow.data.value.ValueChangeMode;
import com.vaadin.flow.server.StreamResource;

public abstract class ApplicationsUI
    extends VerticalLayout
{

    private static final long serialVersionUID = 8127922694035530089L;

    private static final int FILTER_VALUE_CHANGE_TIMEOUT = 500;

    private static final int VISIBLE_ROWS = 50;

    private static final long NO_ELEMENTS = 0;

    private static final String CLASS_GRID_GROUPS = "grid-groups";

    private static final String CLASS_GRID_GROUPS_ICON = "grid-groups-icon";

    private static final String CLASS_GRID_GROUPS_LABEL = "grid-groups-label";

    private static final String CLASS_ICON = "icon";

    private static final String CLASS_SETTINGS = "settings";

    private static final String ICON_NAME_SEPARATOR = ":";

    private static final String NAME_WITH_NOTES = "%s (%s)";

    private static final String STYLE_COLOR = "color";

    private static final String STYLE_COLOR_RED_VALUE = "red";

    private static final String STYLE_DISPLAY = "display";

    private static final String STYLE_DISPLAY_VALUE = "grid";

    private static final String STYLE_GRID_COMBOBOX = "grid-combobox";

    private static final String STYLE_HEIGHT = "height";

    private static final String STYLE_HEIGHT_0PX_VALUE = "0px";

    private static final String STYLE_HEIGHT_16PX_VALUE = "16px";

    private static final String STYLE_MARGIN_LEFT = "margin-left";

    private static final String STYLE_MARGIN_RIGHT = "margin-right";

    private static final String STYLE_MARGIN_VALUE = "1px";

    private static final String STYLE_TEXT_ALIGN = "text-align";

    private static final String STYLE_TEXT_ALIGN_VALUE = "center";

    private static final String STYLE_WIDTH = "width";

    private static final String STYLE_WIDTH_0PX_VALUE = "0px";

    private static final String STYLE_WIDTH_100PX_VALUE = "100px";

    private static final String STYLE_WIDTH_130PX_VALUE = "130px";

    private static final String STYLE_WIDTH_150PX_VALUE = "150px";

    private static final String STYLE_WIDTH_16PX_VALUE = "16px";

    private static final String STYLE_WIDTH_70PX_VALUE = "70px";

    private long userId = UserData.NEW_USER;

    private transient List<ApplicationData> applicationDatas = null;

    private transient Map<ApplicationColumns, String> filterMap = null;

    private Database database = null;

    private DataProvider<ApplicationData, Void> dataProvider = null;

    private Grid<ApplicationData> applicationGrid = null;

    private Grid.Column<ApplicationData> editDeleteColumn = null;

    private VirtualUtil virtualUtil = null;

    private ApplicationEditDialog applicationEditDialog = null;

    private ApplicationsUI()
    {
        setMargin( false );
        setSpacing( false );
        setSizeFull();
        addClassName( CLASS_SETTINGS );
        filterMap = new EnumMap<>( ApplicationColumns.class );
    }

    public ApplicationsUI( long userId, Database database, VirtualUtil virtualUtil )
    {
        this();
        this.userId = userId;
        this.database = database;
        this.virtualUtil = virtualUtil;

        dataProvider = DataProvider.fromCallbacks( query -> {
            int offset = query.getOffset();
            int limit = query.getLimit();

            List<String> orders = new ArrayList<>();
            for ( SortOrder<String> order : query.getSortOrders() )
            {
                orders.add( SortDirection.DESCENDING.equals( order.getDirection() ) ? Database.ORDER_DESCENDING + order.getSorted() : order.getSorted() );
            }

            return new ArrayList<>( this.database.getApplications( offset, limit, filterMap, orders ) ).stream();
        }, query -> database.getApplicationRows( filterMap )
                            .intValue() );

        UI ui = UI.getCurrent();
        applicationDatas = new ArrayList<>( database.getAllApplications() );
        applicationGrid = new Grid<>();
        applicationGrid.setDataProvider( dataProvider );
        applicationGrid.setPageSize( VISIBLE_ROWS );
        applicationGrid.getStyle()
                       .set( STYLE_DISPLAY, STYLE_DISPLAY_VALUE )
                       .set( STYLE_MARGIN_LEFT, STYLE_MARGIN_VALUE )
                       .set( STYLE_MARGIN_RIGHT, STYLE_MARGIN_VALUE );
        applicationGrid.addItemDoubleClickListener( event -> {
            virtualUtil.setApplicationTestingEditDialogOpened( true );
            editApplicationData( ui, event.getItem(), true );
            applicationGrid.select( event.getItem() );
        } );
        add( applicationGrid );

        // icon
        applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationData ) -> {
            IconData iconData = applicationData.getIconData();
            if ( Objects.nonNull( iconData ) )
            {
                if ( Objects.isNull( iconData.getIcon() ) )
                {
                    String iconName = iconData.getName();
                    div.add( new IronIcon( iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[0] : iconName,
                                           iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[1] : iconName ) );
                }
                else
                {
                    StreamResource streamResource = new StreamResource( iconData.getName(), () -> new ByteArrayInputStream( iconData.getIcon() ) );
                    Image image = new Image( streamResource, iconData.getName() );
                    image.addClassName( CLASS_ICON );
                    div.add( image );
                }
            }
        } ) )
                       .setFlexGrow( 0 )
                       .setWidth( STYLE_WIDTH_70PX_VALUE )
                       .setResizable( false )
                       .setHeader( virtualUtil.getTranslation( ApplicationColumns.ICON.getValueForTranslation() ) );

        // name
        Grid.Column<ApplicationData> nameColumn = applicationGrid.addColumn( ApplicationData::getName )
                                                                 .setFlexGrow( 1 )
                                                                 .setResizable( true )
                                                                 .setSortable( true )
                                                                 .setSortProperty( ApplicationColumns.NAME.getName() )
                                                                 .setWidth( STYLE_WIDTH_150PX_VALUE )
                                                                 .setHeader( virtualUtil.getTranslation( ApplicationColumns.NAME.getValueForTranslation() ) );
        // command
        Grid.Column<ApplicationData> commandColumn = applicationGrid.addColumn( ApplicationData::getCommand )
                                                                    .setFlexGrow( 2 )
                                                                    .setResizable( true )
                                                                    .setSortable( true )
                                                                    .setSortProperty( ApplicationColumns.COMMAND.getName() )
                                                                    .setHeader( virtualUtil.getTranslation( ApplicationColumns.COMMAND.getValueForTranslation() ) );
        // command options
        Grid.Column<ApplicationData> commandOptionsColumn = applicationGrid.addColumn( ApplicationData::getCommandOptions )
                                                                           .setFlexGrow( 2 )
                                                                           .setResizable( true )
                                                                           .setSortable( true )
                                                                           .setSortProperty( ApplicationColumns.COMMAND_OPTIONS.getName() )
                                                                           .setHeader( virtualUtil.getTranslation( ApplicationColumns.COMMAND_OPTIONS.getValueForTranslation() ) );
        // note
        Grid.Column<ApplicationData> noteColumn = applicationGrid.addColumn( ApplicationData::getNote )
                                                                 .setFlexGrow( 3 )
                                                                 .setResizable( true )
                                                                 .setSortable( true )
                                                                 .setSortProperty( ApplicationColumns.NOTE.getName() )
                                                                 .setHeader( virtualUtil.getTranslation( ApplicationColumns.NOTE.getValueForTranslation() ) );
        // servers group
        Grid.Column<ApplicationData> serversGroupColumn = applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( serverGroupsDiv, applicationData ) -> {
            List<ServersGroupData> serversGroupDatas = applicationData.getServersGroupDatas();
            serversGroupDatas.forEach( serversGroupData -> {
                Label label = new Label( StringUtils.isBlank( serversGroupData.getNote() ) ? serversGroupData.getName()
                                : String.format( NAME_WITH_NOTES, serversGroupData.getName(), serversGroupData.getNote() ) );
                label.addClassName( CLASS_GRID_GROUPS_LABEL );
                Icon icon = serversGroupData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
                icon.addClassName( CLASS_GRID_GROUPS_ICON );

                Div serversGroupDiv = new Div();
                serversGroupDiv.addClassName( CLASS_GRID_GROUPS );
                serversGroupDiv.add( icon );
                serversGroupDiv.add( label );
                serverGroupsDiv.add( serversGroupDiv );
            } );
        } ) )
                                                                         .setFlexGrow( 3 )
                                                                         .setResizable( true )
                                                                         .setSortable( false )
                                                                         .setHeader( virtualUtil.getTranslation( ApplicationColumns.SERVERS_GROUP.getValueForTranslation() ) );
        // server auto select
        Grid.Column<ApplicationData> serverAutoSelectColumn = applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationData ) -> {
            Icon icon = applicationData.isServerAutoSelect() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                             .setFlexGrow( 0 )
                                                                             .setResizable( true )
                                                                             .setSortable( true )
                                                                             .setSortProperty( ApplicationColumns.SERVER_AUTO_SELECT.getName() )
                                                                             .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                                             .setHeader( virtualUtil.getTranslation( ApplicationColumns.SERVER_AUTO_SELECT.getValueForTranslation() ) );
        // one instance
        Grid.Column<ApplicationData> oneInstanceColumn = applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationData ) -> {
            Icon icon = applicationData.isOneInstance() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                        .setFlexGrow( 0 )
                                                                        .setResizable( true )
                                                                        .setSortable( true )
                                                                        .setSortProperty( ApplicationColumns.ONE_INSTANCE.getName() )
                                                                        .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                                        .setHeader( virtualUtil.getTranslation( ApplicationColumns.ONE_INSTANCE.getValueForTranslation() ) );
        // enabled
        Grid.Column<ApplicationData> enabledColumn = applicationGrid.addColumn( new ComponentRenderer<>( Div::new, ( div, applicationData ) -> {
            Icon icon = applicationData.isEnabled() ? new Icon( VaadinIcon.CHECK ) : new Icon( VaadinIcon.CLOSE );
            icon.getStyle()
                .set( STYLE_HEIGHT, STYLE_HEIGHT_16PX_VALUE )
                .set( STYLE_WIDTH, STYLE_WIDTH_16PX_VALUE );
            div.add( icon );
            div.getStyle()
               .set( STYLE_TEXT_ALIGN, STYLE_TEXT_ALIGN_VALUE );
        } ) )
                                                                    .setFlexGrow( 0 )
                                                                    .setResizable( true )
                                                                    .setSortable( true )
                                                                    .setSortProperty( ApplicationColumns.ENABLED.getName() )
                                                                    .setWidth( STYLE_WIDTH_130PX_VALUE )
                                                                    .setHeader( virtualUtil.getTranslation( ApplicationColumns.ENABLED.getValueForTranslation() ) );
        // edit/delete data
        editDeleteColumn = applicationGrid.addColumn( new ComponentRenderer<>( applicationData -> {
            // delete
            Button deleteButton = new Button( new Icon( VaadinIcon.TRASH ), event -> {
                boolean noElements = ( database.getApplicationsGroupApplicationRows( applicationData ) + database.getUserApplicationRows( applicationData ) ) == NO_ELEMENTS;
                String message = noElements ? virtualUtil.getTranslation( Translation.APPLICATION_REMOVE, applicationData.getName() )
                                : virtualUtil.getTranslation( Translation.APPLICATION_REMOVE_APPLICATIONS_GROUP, applicationData.getName() );
                ApplicationRemoveDialog dialog = new ApplicationRemoveDialog( message, virtualUtil )
                {
                    private static final long serialVersionUID = 4567922694035530089L;

                    @Override
                    public void savingData()
                    {
                        removeApplication( applicationData );
                    }
                };
                dialog.open();
                applicationGrid.select( applicationData );
            } );
            deleteButton.getStyle()
                        .set( STYLE_COLOR, STYLE_COLOR_RED_VALUE );

            // edit
            Button editButton = new Button( new Icon( VaadinIcon.EDIT ), event -> {
                virtualUtil.setApplicationTestingEditDialogOpened( true );
                editApplicationData( ui, applicationData, true );
                applicationGrid.select( applicationData );
            } );

            if ( virtualUtil.isTooltipsVisible() )
            {
                TooltipRunner tooltipRunner1 = new TooltipRunner( deleteButton, virtualUtil.getTranslation( Translation.BUTTON_DELETE ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner1 );
                TooltipRunner tooltipRunner2 = new TooltipRunner( editButton, virtualUtil.getTranslation( Translation.BUTTON_EDIT ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner2 );
            }

            HorizontalLayout layout = new HorizontalLayout( editButton, deleteButton );
            layout.setMargin( false );
            layout.setPadding( false );
            layout.setSpacing( false );
            layout.setSizeUndefined();
            return layout;
        } ) )
                                          .setFlexGrow( 0 )
                                          .setResizable( false )
                                          .setWidth( STYLE_WIDTH_100PX_VALUE );
        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationDatas ) );

        // filter: name
        HeaderRow filterRow = applicationGrid.appendHeaderRow();
        TextField nameFilter = new TextField();
        nameFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        nameFilter.setSizeFull();
        nameFilter.setValueChangeMode( ValueChangeMode.LAZY );
        nameFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        nameFilter.addValueChangeListener( event -> {
            if ( nameFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.NAME );
            }
            else
            {
                filterMap.put( ApplicationColumns.NAME, nameFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( nameColumn )
                 .setComponent( nameFilter );

        // filter: command
        TextField commandFilter = new TextField();
        commandFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        commandFilter.setSizeFull();
        commandFilter.setValueChangeMode( ValueChangeMode.LAZY );
        commandFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        commandFilter.addValueChangeListener( event -> {
            if ( commandFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.COMMAND );
            }
            else
            {
                filterMap.put( ApplicationColumns.COMMAND, commandFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( commandColumn )
                 .setComponent( commandFilter );

        // filter: command options
        TextField commandOptionsFilter = new TextField();
        commandOptionsFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        commandOptionsFilter.setSizeFull();
        commandOptionsFilter.setValueChangeMode( ValueChangeMode.LAZY );
        commandOptionsFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        commandOptionsFilter.addValueChangeListener( event -> {
            if ( commandFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.COMMAND_OPTIONS );
            }
            else
            {
                filterMap.put( ApplicationColumns.COMMAND_OPTIONS, commandFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( commandOptionsColumn )
                 .setComponent( commandOptionsFilter );

        // filter: note
        TextField noteFilter = new TextField();
        noteFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        noteFilter.setSizeFull();
        noteFilter.setValueChangeMode( ValueChangeMode.LAZY );
        noteFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        noteFilter.addValueChangeListener( event -> {
            if ( noteFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.NOTE );
            }
            else
            {
                filterMap.put( ApplicationColumns.NOTE, noteFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( noteColumn )
                 .setComponent( noteFilter );

        // filter: servers group
        TextField serversGroupFilter = new TextField();
        serversGroupFilter.setPlaceholder( virtualUtil.getTranslation( Translation.FIELD_FILTER ) );
        serversGroupFilter.setSizeFull();
        serversGroupFilter.setValueChangeMode( ValueChangeMode.LAZY );
        serversGroupFilter.setValueChangeTimeout( FILTER_VALUE_CHANGE_TIMEOUT );
        serversGroupFilter.addValueChangeListener( event -> {
            if ( serversGroupFilter.isEmpty() )
            {
                filterMap.remove( ApplicationColumns.SERVERS_GROUP );
            }
            else
            {
                filterMap.put( ApplicationColumns.SERVERS_GROUP, serversGroupFilter.getValue() );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( serversGroupColumn )
                 .setComponent( serversGroupFilter );

        // filter: server auto select
        String filterAll = virtualUtil.getTranslation( Translation.DATA_FILTER_ALL );
        String filterNo = virtualUtil.getTranslation( Translation.DATA_FILTER_NO );
        String filterYes = virtualUtil.getTranslation( Translation.DATA_FILTER_YES );
        ComboBox<String> serverAutoSelectFilter = new ComboBox<>();
        serverAutoSelectFilter.setAllowCustomValue( false );
        serverAutoSelectFilter.setClassName( STYLE_GRID_COMBOBOX );
        serverAutoSelectFilter.setItems( filterAll, filterYes, filterNo );
        serverAutoSelectFilter.setValue( filterAll );
        serverAutoSelectFilter.setSizeFull();
        serverAutoSelectFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( serverAutoSelectFilter.getValue(), filterYes ) )
            {
                filterMap.put( ApplicationColumns.SERVER_AUTO_SELECT, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( serverAutoSelectFilter.getValue(), filterNo ) )
            {
                filterMap.put( ApplicationColumns.SERVER_AUTO_SELECT, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ApplicationColumns.SERVER_AUTO_SELECT );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( serverAutoSelectColumn )
                 .setComponent( serverAutoSelectFilter );

        // filter: one instance
        ComboBox<String> oneInstanceFilter = new ComboBox<>();
        oneInstanceFilter.setAllowCustomValue( false );
        oneInstanceFilter.setClassName( STYLE_GRID_COMBOBOX );
        oneInstanceFilter.setItems( filterAll, filterYes, filterNo );
        oneInstanceFilter.setValue( filterAll );
        oneInstanceFilter.setSizeFull();
        oneInstanceFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( oneInstanceFilter.getValue(), filterYes ) )
            {
                filterMap.put( ApplicationColumns.ONE_INSTANCE, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( oneInstanceFilter.getValue(), filterNo ) )
            {
                filterMap.put( ApplicationColumns.ONE_INSTANCE, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ApplicationColumns.ONE_INSTANCE );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( oneInstanceColumn )
                 .setComponent( oneInstanceFilter );

        // filter: enabled
        ComboBox<String> enabledFilter = new ComboBox<>();
        enabledFilter.setAllowCustomValue( false );
        enabledFilter.setClassName( STYLE_GRID_COMBOBOX );
        enabledFilter.setItems( filterAll, filterYes, filterNo );
        enabledFilter.setValue( filterAll );
        enabledFilter.setSizeFull();
        enabledFilter.addValueChangeListener( event -> {
            if ( StringUtils.equals( enabledFilter.getValue(), filterYes ) )
            {
                filterMap.put( ApplicationColumns.ENABLED, Boolean.TRUE.toString() );
            }
            else if ( StringUtils.equals( enabledFilter.getValue(), filterNo ) )
            {
                filterMap.put( ApplicationColumns.ENABLED, Boolean.FALSE.toString() );
            }
            else
            {
                filterMap.remove( ApplicationColumns.ENABLED );
            }
            dataProvider.refreshAll();
        } );
        filterRow.getCell( enabledColumn )
                 .setComponent( enabledFilter );
    }

    public abstract void testApplication( ApplicationData applicationData );

    public void editApplicationData( UI ui, ApplicationData applicationData, boolean testingApplicationEnabled )
    {
        if ( virtualUtil.isApplicationTestingEditDialogOpened() || virtualUtil.isNotApplicationTestingError() )
        {
            applicationEditDialog = new ApplicationEditDialog( ui, userId, applicationData, database, testingApplicationEnabled, virtualUtil )
            {
                private static final long serialVersionUID = -2127922694035530089L;

                @Override
                public void applicationTest( ApplicationData applicationData )
                {
                    virtualUtil.setApplicationTestingError( false );
                    testApplication( applicationData );
                    close();
                }

                @Override
                public void savingData( ApplicationData data )
                {
                    if ( Objects.nonNull( data ) )
                    {
                        boolean closeDialog = true;
                        if ( Objects.nonNull( database.getApplication( data.getId() ) ) )
                        {
                            closeDialog = updateApplication( data );
                        }
                        else
                        {
                            closeDialog = addApplication( data );
                        }
                        if ( closeDialog )
                        {
                            // DON'T REMOVE
                            setWidth( STYLE_WIDTH_0PX_VALUE );
                            setHeight( STYLE_HEIGHT_0PX_VALUE );
                            close();
                        }
                    }
                    else
                    {
                        new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                    }
                }
            };
            applicationEditDialog.open();
        }
    }

    public ApplicationEditDialog getApplicationEditDialog()
    {
        return applicationEditDialog;
    }

    public boolean addApplication( ApplicationData applicationData )
    {
        if ( Objects.nonNull( applicationData ) )
        {
            applicationData.setId( database.addApplication( applicationData ) );
            if ( !Objects.equals( applicationData.getId(), ApplicationData.UNDEFINED_APPLICATION ) )
            {
                applicationDatas.add( applicationData );
                dataProvider.refreshAll();
                applicationGrid.select( applicationData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_ADD ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean removeApplication( ApplicationData applicationData )
    {
        if ( Objects.nonNull( applicationData ) )
        {
            if ( database.removeApplication( applicationData ) )
            {
                applicationGrid.deselectAll();
                applicationDatas.remove( applicationData );
                dataProvider.refreshAll();
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_REMOVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        else
        {
            new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
        }
        return false;
    }

    public boolean updateApplication( ApplicationData applicationData )
    {
        if ( Objects.nonNull( applicationData ) )
        {
            if ( database.updateApplication( applicationData ) )
            {
                applicationDatas.remove( applicationData );
                applicationDatas.add( applicationData );
                dataProvider.refreshAll();
                applicationGrid.select( applicationData );
                editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationDatas ) );
                return true;
            }
            else
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE, applicationData.getName() ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        return false;
    }

    public void addNewApplication()
    {
        ApplicationData applicationData = new ApplicationData();
        applicationData.setId( ApplicationData.NEW_APPLICATION );
        applicationEditDialog = new ApplicationEditDialog( UI.getCurrent(), userId, applicationData, database, true, virtualUtil )
        {
            private static final long serialVersionUID = -9172182851699419980L;

            @Override
            public void applicationTest( ApplicationData applicationData )
            {
                if ( StringUtils.isNotBlank( applicationData.getCommand() ) && CollectionUtils.isNotEmpty( applicationData.getServersGroupDatas() ) )
                {
                    virtualUtil.setApplicationTestingError( false );
                    virtualUtil.setApplicationTestingEditDialogOpened( true );
                    testApplication( applicationData );
                    close();
                }
            }

            @Override
            public void savingData( ApplicationData applicationData )
            {
                if ( Objects.nonNull( applicationData ) )
                {
                    applicationData.setId( database.addApplication( applicationData ) );
                    if ( !Objects.equals( applicationData.getId(), ApplicationData.UNDEFINED_APPLICATION ) )
                    {
                        applicationDatas.add( applicationData );
                        dataProvider.refreshAll();
                        applicationGrid.select( applicationData );
                        editDeleteColumn.setVisible( CollectionUtils.isNotEmpty( applicationDatas ) );
                        close();
                    }
                    else
                    {
                        new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_EXISTS, applicationData.getName() ),
                                            virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                    }
                }
                else
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_SAVE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
            }
        };
        applicationEditDialog.open();
    }

    public void selectApplication( ApplicationData applicationData )
    {
        if ( Objects.nonNull( applicationData ) )
        {
            applicationGrid.select( applicationData );
        }
    }

}
