/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.bars.running;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.icons.JqueryuiIcon;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;

import com.vaadin.flow.component.Component;
import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Image;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.IronIcon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.tabs.Tab;
import com.vaadin.flow.component.tabs.Tabs;
import com.vaadin.flow.server.StreamResource;

import dev.mett.vaadin.tooltip.Tooltips;

public abstract class RunningApplicationsBar
    extends HorizontalLayout
{

    private static final long serialVersionUID = -826623092527280598L;

    public static final long NO_RUNNING_APPLICATIONS = 0;

    private static final long ONLY_ONE_RUNNING_APPLICATION = 1;

    private static final int UNSELECTED_TAB = -1;

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "small";

    private static final String CLASS_ICON = "icon";

    private static final String CLASS_ICON_CLOSE = "close";

    private static final String CLASS_RUNNING_APPLICATION = "running-application";

    private static final String CLASS_RUNNING_APPLICATIONS_BAR = "running-applications-bar";

    private static final String ICON_NAME_SEPARATOR = ":";

    private boolean selectedTabChangedAction = true;

    private long oldRunningApplicationId = NO_RUNNING_APPLICATIONS;

    private String tooltipContext = null;

    private transient List<Div> closeDivs = null;

    private VirtualUtil virtualUtil = null;

    private Tabs tabs = null;

    public RunningApplicationsBar( VirtualUtil virtualUtil )
    {
        setJustifyContentMode( JustifyContentMode.CENTER );
        setMargin( false );
        setPadding( false );
        setSpacing( false );
        setSizeUndefined();
        addClassNames( CLASS_RUNNING_APPLICATIONS_BAR );
        closeDivs = new ArrayList<>();
        this.virtualUtil = virtualUtil;

        // icon
        Icon icon = new Icon( VaadinIcon.DESKTOP );
        setVerticalComponentAlignment( Alignment.CENTER, icon );
        add( icon );

        // tabs
        tabs = new Tabs();
        setVerticalComponentAlignment( Alignment.CENTER, tabs );
        tabs.getElement()
            .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        tabs.addSelectedChangeListener( event -> {
            Tab tab = tabs.getSelectedTab();
            if ( Objects.nonNull( tab ) )
            {
                if ( selectedTabChangedAction )
                {
                    closeOldAndOpenNewRunningApplication();
                }

                if ( virtualUtil.isTooltipsVisible() && !virtualUtil.isAllRunningInSameInstance() && isNotOnlyOneRunningApplication() )
                {
                    Tab previousTab = event.getPreviousTab();
                    if ( Objects.nonNull( previousTab ) && StringUtils.isNotBlank( tooltipContext ) )
                    {
                        TooltipRunner tooltipRunner = new TooltipRunner( previousTab, tooltipContext, UI.getCurrent() );
                        virtualUtil.startRunnerService( tooltipRunner );
                        tooltipContext = StringUtils.EMPTY;
                    }
                }

                // set only button of active tab visible
                closeDivs.stream()
                         .filter( Objects::nonNull )
                         .forEach( div -> div.setVisible( tabs.getSelectedIndex() != UNSELECTED_TAB && div.hasClassName( tab.getId()
                                                                                                                            .orElse( StringUtils.EMPTY ) ) ) );
                if ( virtualUtil.isTooltipsVisible() && !virtualUtil.isAllRunningInSameInstance() )
                {
                    Tooltips.getCurrent()
                            .removeTooltip( tab );
                }
            }
            selectedTabChangedAction = true;
        } );
        add( tabs );
    }

    public abstract void closeRunningApplication();

    public abstract void closeOldAndOpenNewRunningApplication();

    private Tab findTab( Long tabId )
    {
        Tab tab = null;
        for ( int index = 0; index < tabs.getComponentCount(); index++ )
        {
            if ( StringUtils.equals( ( (Tab) tabs.getComponentAt( index ) ).getId()
                                                                           .orElse( StringUtils.EMPTY ),
                                     tabId.toString() ) )
            {
                tab = (Tab) tabs.getComponentAt( index );
                break;
            }
        }
        return tab;
    }

    public boolean runningApplicationNotExists( Long runningApplicationId )
    {
        return ( Objects.isNull( findTab( runningApplicationId ) ) );
    }

    public boolean isEmpty()
    {
        return ( tabs.getComponentCount() == NO_RUNNING_APPLICATIONS );
    }

    public boolean isNotEmpty()
    {
        return !isEmpty();
    }

    public boolean isOnlyOneRunningApplication()
    {
        return ( tabs.getComponentCount() == ONLY_ONE_RUNNING_APPLICATION );
    }

    public boolean isNotOnlyOneRunningApplication()
    {
        return !isOnlyOneRunningApplication();
    }

    public Long getActiveRunningApplicationId()
    {
        if ( isNotEmpty() )
        {
            String id = StringUtils.EMPTY;
            if ( Objects.nonNull( tabs.getSelectedTab() ) )
            {
                id = tabs.getSelectedTab()
                         .getId()
                         .orElse( StringUtils.EMPTY );
            }
            return ( StringUtils.equals( StringUtils.EMPTY, id ) ? NO_RUNNING_APPLICATIONS : Long.parseLong( id ) );
        }
        return NO_RUNNING_APPLICATIONS;
    }

    public void setRunningApplicationActive( long tabId )
    {
        if ( isNotEmpty() )
        {
            Tab tab = findTab( tabId );
            if ( Objects.nonNull( tab ) )
            {
                selectedTabChangedAction = true;
                tabs.setSelectedTab( tab );
            }
        }
    }

    public void setRunningApplicationInactive()
    {
        if ( isNotEmpty() )
        {
            selectedTabChangedAction = false;
            tabs.setSelectedIndex( UNSELECTED_TAB );
            closeDivs.stream()
                     .filter( Objects::nonNull )
                     .forEach( div -> div.setVisible( false ) );
        }
    }

    public long getOldRunningApplicationId()
    {
        return oldRunningApplicationId;
    }

    public void setOldRunningApplicationId( long oldRunningApplicationId )
    {
        this.oldRunningApplicationId = oldRunningApplicationId;
    }

    public boolean addRunningApplication( Long runningApplicationId, String title, IconData iconData )
    {
        boolean runningApplicationAdded = false;
        if ( runningApplicationNotExists( runningApplicationId ) && Objects.nonNull( iconData ) )
        {
            Div div = new Div();

            // icon
            if ( Objects.isNull( iconData.getIcon() ) )
            {
                String iconName = iconData.getName();
                IronIcon ironIcon = new IronIcon( iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[0] : iconName,
                                                  iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[1] : iconName );
                ironIcon.addClassName( CLASS_ICON );
                div.add( ironIcon );
            }
            else
            {
                StreamResource streamResource = new StreamResource( iconData.getName(), () -> new ByteArrayInputStream( iconData.getIcon() ) );
                Image image = new Image( streamResource, iconData.getName() );
                image.addClassName( CLASS_ICON );
                div.add( image );
            }

            // label
            Label label = new Label( title );
            div.add( label );

            // close icon
            Icon closeIcon = JqueryuiIcon.UI_ICON_CLOSETHICK.create();
            Div closeDiv = new Div( closeIcon );
            closeDiv.setVisible( false );
            closeDiv.addClassNames( CLASS_ICON_CLOSE, runningApplicationId.toString() );
            closeDiv.addClickListener( event -> {
                CloseApplicationDialog closeApplicationDialog = new CloseApplicationDialog();
                closeApplicationDialog.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_CLOSE ) );
                closeApplicationDialog.setCancelButtonText( virtualUtil.getTranslation( Translation.BUTTON_CANCEL ) );
                closeApplicationDialog.setCloseButtonText( virtualUtil.getTranslation( Translation.BUTTON_APPLICATION_CLOSE ) );
                closeApplicationDialog.setCloseButtonClickListener( e -> {
                    closeApplicationDialog.close();
                    closeRunningApplication();
                } );
                closeApplicationDialog.open();
            } );
            div.add( closeDiv );
            closeDivs.add( closeDiv );

            // tab
            selectedTabChangedAction = false;
            setOldRunningApplicationId( runningApplicationId );
            Tab tab = new Tab( div );
            tab.addClassName( CLASS_RUNNING_APPLICATION );
            tab.setId( runningApplicationId.toString() );
            tabs.add( tab );
            tabs.setSelectedTab( tab );
            runningApplicationAdded = true;

            if ( virtualUtil.isTooltipsVisible() && !virtualUtil.isAllRunningInSameInstance() )
            {
                TooltipRunner tooltipRunner1 = new TooltipRunner( closeDiv, virtualUtil.getTranslation( Translation.BUTTON_APPLICATION_CLOSE ), UI.getCurrent() );
                virtualUtil.startRunnerService( tooltipRunner1 );

                // set tab tooltip
                if ( StringUtils.isNotBlank( tooltipContext ) )
                {
                    TooltipRunner tooltipRunner2 = new TooltipRunner( tab, tooltipContext, UI.getCurrent() );
                    virtualUtil.startRunnerService( tooltipRunner2 );
                }
            }
        }
        return runningApplicationAdded;
    }

    public void removeAllRunningApplications()
    {
        closeDivs.forEach( div -> Tooltips.getCurrent()
                                          .removeTooltip( div ) );
        setOldRunningApplicationId( NO_RUNNING_APPLICATIONS );
        selectedTabChangedAction = false;
        closeDivs.clear();
        tabs.removeAll();
    }

    public void removeRunningApplication( Long runningApplicationId )
    {
        Tab tab = findTab( runningApplicationId );
        if ( Objects.nonNull( tab ) )
        {
            selectedTabChangedAction = true;
            setOldRunningApplicationId( NO_RUNNING_APPLICATIONS );
            tabs.remove( tab );

            // remove deleted div from divs list
            closeDivs.stream()
                     .filter( div -> div.hasClassName( Long.toString( runningApplicationId ) ) )
                     .collect( Collectors.toList() )
                     .forEach( div -> {
                         closeDivs.remove( div );
                         Tooltips.getCurrent()
                                 .removeTooltip( div );
                     } );
        }
    }

    public void updateTitle( String title, IconData iconData )
    {
        Tab tab = findTab( getActiveRunningApplicationId() );
        if ( Objects.nonNull( tab ) )
        {
            Component component = tab.getChildren()
                                     .findFirst()
                                     .orElse( null );
            if ( Objects.nonNull( component ) )
            {
                component.getChildren()
                         .forEach( c -> {
                             if ( c instanceof Label )
                             {
                                 Label label = (Label) c;
                                 label.setText( title );
                             }
                             else if ( c instanceof Image || c instanceof IronIcon )
                             {
                                 Div div = (Div) component;
                                 div.remove( c );
                                 if ( Objects.isNull( iconData.getIcon() ) )
                                 {
                                     String iconName = iconData.getName();
                                     IronIcon ironIcon = new IronIcon( iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[0] : iconName,
                                                                       iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[1] : iconName );
                                     ironIcon.addClassName( CLASS_ICON );
                                     div.addComponentAsFirst( ironIcon );
                                 }
                                 else
                                 {
                                     StreamResource streamResource = new StreamResource( iconData.getName(), () -> new ByteArrayInputStream( iconData.getIcon() ) );
                                     Image image = new Image( streamResource, iconData.getName() );
                                     image.addClassName( CLASS_ICON );
                                     div.addComponentAsFirst( image );
                                 }
                             }
                         } );
            }
        }

    }

    public void setTooltipContext( String tooltipContext )
    {
        this.tooltipContext = tooltipContext;
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        closeDivs.forEach( Div::removeAll );
        closeDivs.clear();
    }

}
