/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.bars.running;

import java.io.ByteArrayInputStream;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.icons.JqueryuiIcon;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.xpra.packets.base.WindowType;

import com.vaadin.flow.component.Component;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Image;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.tabs.Tab;
import com.vaadin.flow.component.tabs.Tabs;
import com.vaadin.flow.server.StreamResource;

import dev.mett.vaadin.tooltip.Tooltips;

public abstract class WindowsBar
    extends HorizontalLayout
{

    private static final long serialVersionUID = 1573860901284763189L;

    public static final int NO_WINDOWS = 0;

    public static final int ONLY_ONE_WINDOW = 1;

    private static final int TITLE_MAX_WIDTH = 50;

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "small";

    private static final String CLASS_DISABLED = "disabled";

    private static final String CLASS_DISABLED_SELECTED = "selected";

    private static final String CLASS_ICON = "icon";

    private static final String CLASS_ICON_CLOSE = "close";

    private static final String CLASS_ICON_MAXIMIZE = "maximize";

    private static final String CLASS_ICON_RESTORE = "restore";

    private static final String CLASS_WINDOWS_BAR = "windows-bar";

    private boolean windowsVisible = false;

    private String selectedTabId = StringUtils.EMPTY;

    private VirtualUtil virtualUtil = null;

    private Tabs tabs = null;

    public enum WindowsBarAction
    {
        CLOSE_WINDOW, MAXIMIZE_WINDOW, NEW_WINDOW_FOCUS, RESTORE_WINDOW
    }

    public abstract void windowsBarAction( int id, WindowsBarAction action );

    protected WindowsBar( VirtualUtil virtualUtil )
    {
        super();
        setMargin( false );
        setPadding( false );
        setSpacing( false );
        setSizeUndefined();
        addClassName( CLASS_WINDOWS_BAR );
        this.virtualUtil = virtualUtil;

        // icon
        Icon icon = new Icon( virtualUtil.isAllRunningInSameInstance() ? VaadinIcon.DESKTOP : VaadinIcon.LAPTOP );
        setVerticalComponentAlignment( Alignment.CENTER, icon );
        add( icon );

        // tabs
        tabs = new Tabs();
        setVerticalComponentAlignment( Alignment.CENTER, tabs );
        tabs.getElement()
            .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        tabs.addSelectedChangeListener( event -> {
            String newSelectedTabId = getSelectedTabId();
            if ( !StringUtils.equals( newSelectedTabId, Integer.toString( NO_WINDOWS ) ) )
            {
                windowsBarAction( Integer.parseInt( newSelectedTabId ), WindowsBarAction.NEW_WINDOW_FOCUS );
                setNewWindowFocus( Integer.parseInt( newSelectedTabId ) );
            }
        } );
        add( tabs );
    }

    private Tab createTab( int id, String title, byte[] icon )
    {
        Div div = new Div();

        if ( Objects.nonNull( icon ) && icon.length > 0 )
        {
            StreamResource streamResource = new StreamResource( StringUtils.EMPTY, () -> new ByteArrayInputStream( icon ) );
            Image image = new Image( streamResource, StringUtils.EMPTY );
            image.addClassName( CLASS_ICON );
            div.add( image );
        }

        Label label = new Label( StringUtils.abbreviate( title, TITLE_MAX_WIDTH ) );
        div.add( label );
        if ( virtualUtil.isTooltipsVisible() && StringUtils.length( title ) >= TITLE_MAX_WIDTH )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( label, title, UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Icon restoreIcon = JqueryuiIcon.UI_ICON_WINDOWS.create();
        Div restoreDiv = new Div( restoreIcon );
        restoreDiv.addClassNames( CLASS_ICON_RESTORE, CLASS_DISABLED );
        restoreDiv.addClickListener( event -> windowsBarAction( id, WindowsBarAction.RESTORE_WINDOW ) );
        div.add( restoreDiv );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( restoreDiv, virtualUtil.getTranslation( Translation.BUTTON_WINDOW_RESTORE ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Icon maximizeIcon = JqueryuiIcon.UI_ICON_ACTION.create();
        Div maximizeDiv = new Div( maximizeIcon );
        maximizeDiv.addClassName( CLASS_ICON_MAXIMIZE );
        maximizeDiv.addClickListener( event -> windowsBarAction( id, WindowsBarAction.MAXIMIZE_WINDOW ) );
        div.add( maximizeDiv );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( maximizeDiv, virtualUtil.getTranslation( Translation.BUTTON_WINDOW_MAXIMIZE ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Icon closeIcon = JqueryuiIcon.UI_ICON_CLOSETHICK.create();
        Div closeDiv = new Div( closeIcon );
        closeDiv.addClassName( CLASS_ICON_CLOSE );
        closeDiv.addClickListener( event -> windowsBarAction( id, WindowsBarAction.CLOSE_WINDOW ) );
        div.add( closeDiv );
        if ( virtualUtil.isTooltipsVisible() )
        {
            TooltipRunner tooltipRunner = new TooltipRunner( closeDiv, virtualUtil.getTranslation( Translation.BUTTON_WINDOW_CLOSE ), UI.getCurrent() );
            virtualUtil.startRunnerService( tooltipRunner );
        }

        Tab tab = new Tab( div );
        tab.setId( Integer.toString( id ) );
        tab.setVisible( windowsVisible );
        selectedTabId = Integer.toString( id );
        return tab;
    }

    private void setButtonDisabled( Tab tab, String iconDivName, boolean disabled )
    {
        Div div = (Div) tab.getChildren()
                           .filter( Objects::nonNull )
                           .findFirst()
                           .orElse( null );
        if ( Objects.nonNull( div ) )
        {
            div.getChildren()
               .filter( Objects::nonNull )
               .forEach( component -> {
                   if ( component instanceof Div )
                   {
                       Div iconDiv = (Div) component;
                       if ( iconDiv.hasClassName( iconDivName ) )
                       {
                           if ( disabled )
                           {
                               iconDiv.removeClassName( CLASS_DISABLED );
                           }
                           else
                           {
                               iconDiv.addClassName( CLASS_DISABLED );
                           }
                       }
                   }
               } );
        }
    }

    private void setDisabledButtonSelected( Tab tab, boolean selected )
    {
        Div div = (Div) tab.getChildren()
                           .findFirst()
                           .orElse( null );
        if ( Objects.nonNull( div ) )
        {
            div.getChildren()
               .filter( Objects::nonNull )
               .forEach( component -> {
                   if ( component instanceof Div )
                   {
                       Div iconDiv = (Div) component;
                       if ( iconDiv.hasClassName( CLASS_DISABLED ) )
                       {
                           if ( selected )
                           {
                               iconDiv.addClassName( CLASS_DISABLED_SELECTED );
                           }
                           else
                           {
                               iconDiv.removeClassName( CLASS_DISABLED_SELECTED );
                           }
                       }
                   }
               } );
        }
    }

    private Tab findTab( int id )
    {
        Tab tab = null;
        for ( int index = 0; index < tabs.getComponentCount(); index++ )
        {
            String tempId = ( (Tab) tabs.getComponentAt( index ) ).getId()
                                                                  .orElse( StringUtils.EMPTY );
            if ( StringUtils.equals( tempId, Integer.toString( id ) ) )
            {
                tab = (Tab) tabs.getComponentAt( index );
                break;
            }
        }
        return tab;
    }

    private String getSelectedTabId()
    {
        Tab selectedTab = tabs.getSelectedTab();
        if ( Objects.nonNull( selectedTab ) && selectedTab.isVisible() )
        {
            return selectedTab.getId()
                              .orElse( Integer.toString( NO_WINDOWS ) );
        }
        return Integer.toString( NO_WINDOWS );
    }

    public boolean isEmpty()
    {
        return ( tabs.getComponentCount() == NO_WINDOWS );
    }

    public boolean isNotEmpty()
    {
        return !isEmpty();
    }

    public boolean isOnlyOneWindow()
    {
        return ( tabs.getComponentCount() == ONLY_ONE_WINDOW );
    }

    public boolean isNotOnlyOneWindow()
    {
        return !isOnlyOneWindow();
    }

    public boolean windowExists( int id )
    {
        if ( !isEmpty() )
        {
            return ( Objects.nonNull( findTab( id ) ) );
        }
        return false;
    }

    public void setAllWindowsVisible()
    {
        if ( !windowsVisible )
        {
            windowsVisible = true;
            for ( int index = 0; index < tabs.getComponentCount(); index++ )
            {
                Tab tab = (Tab) tabs.getComponentAt( index );
                if ( Objects.nonNull( tab ) )
                {
                    tab.setVisible( true );
                }
            }
        }
    }

    public void closeAllWindows()
    {
        for ( int index = 0; index < tabs.getComponentCount(); index++ )
        {
            Tab tab = (Tab) tabs.getComponentAt( index );
            Div div = (Div) tab.getChildren()
                               .findFirst()
                               .orElse( null );
            if ( Objects.nonNull( div ) )
            {
                Component component = div.getChildren()
                                         .findFirst()
                                         .orElse( null );
                if ( component instanceof Image )
                {
                    div.getChildren()
                       .filter( comp -> comp instanceof Image || comp instanceof Label )
                       .forEach( comp -> {
                           if ( comp instanceof Label )
                           {
                               Tooltips.getCurrent()
                                       .removeTooltip( (Label) comp );
                           }
                       } );
                }
            }
            Tooltips.getCurrent()
                    .removeTooltip( tab );
        }
        selectedTabId = StringUtils.EMPTY;
        windowsVisible = false;
        tabs.removeAll();
    }

    public boolean closeWindow( int id )
    {
        boolean closed = false;
        Tab tab = findTab( id );
        if ( Objects.nonNull( tab ) )
        {
            Div div = (Div) tab.getChildren()
                               .findFirst()
                               .orElse( null );
            if ( Objects.nonNull( div ) )
            {
                Component component = div.getChildren()
                                         .findFirst()
                                         .orElse( null );
                if ( component instanceof Image )
                {
                    div.getChildren()
                       .filter( comp -> comp instanceof Image || comp instanceof Label )
                       .forEach( comp -> {
                           if ( comp instanceof Label )
                           {
                               Tooltips.getCurrent()
                                       .removeTooltip( (Label) comp );
                           }
                       } );
                }
            }
            tabs.remove( tab );
            closed = true;
            Tooltips.getCurrent()
                    .removeTooltip( tab );
        }
        return closed;
    }

    public boolean createNewWindow( int id, String title, WindowType windowType, byte[] icon, boolean maximized )
    {
        boolean created = false;
        switch ( windowType )
        {
            case DEFAULT:
            case NORMAL:
            case UTILITY:
                tabs.add( createTab( id, title, icon ) );
                setWindowMaximized( id, maximized );
                created = true;
                break;
            default:
                // nothing to do
                break;
        }
        return created;
    }

    public boolean setNewWindowFocus( int id )
    {
        boolean focused = false;
        Tab tab = null;

        // unselected tab
        for ( int index = 0; index < tabs.getComponentCount(); index++ )
        {
            tab = (Tab) tabs.getComponentAt( index );
            if ( Objects.nonNull( tab ) )
            {
                setDisabledButtonSelected( tab, false );
            }
        }

        // selected tab
        tab = findTab( id );
        if ( Objects.nonNull( tab ) )
        {
            tabs.setSelectedTab( tab );
            setDisabledButtonSelected( tab, true );
            selectedTabId = getSelectedTabId();
            focused = true;
        }
        else
        {
            selectedTabId = StringUtils.EMPTY;
        }

        return focused;
    }

    public boolean setWindowMaximized( int id, boolean maximized )
    {
        boolean maximize = false;
        Tab tab = findTab( id );
        if ( Objects.nonNull( tab ) )
        {
            setButtonDisabled( tab, CLASS_ICON_MAXIMIZE, !maximized );
            setButtonDisabled( tab, CLASS_ICON_RESTORE, maximized );
            setDisabledButtonSelected( tab, StringUtils.equals( selectedTabId, Integer.toString( id ) ) );
            maximize = true;
        }
        return maximize;
    }

    public boolean updateWindow( int id, String title, byte[] icon )
    {
        boolean updated = false;
        Tab tab = findTab( id );
        if ( Objects.nonNull( tab ) )
        {
            boolean tabSelected = tab.isSelected();
            Div div = (Div) tab.getChildren()
                               .findFirst()
                               .orElse( null );
            if ( Objects.nonNull( div ) )
            {
                Component component = div.getChildren()
                                         .findFirst()
                                         .orElse( null );
                if ( component instanceof Image )
                {
                    div.getChildren()
                       .filter( comp -> comp instanceof Image || comp instanceof Label )
                       .forEach( comp -> {
                           if ( comp instanceof Image )
                           {
                               StreamResource streamResource = new StreamResource( StringUtils.EMPTY, () -> new ByteArrayInputStream( icon ) );
                               Image image = (Image) comp;
                               image.setSrc( streamResource );
                           }
                           else if ( comp instanceof Label && StringUtils.isNotBlank( title ) )
                           {
                               Label label = (Label) comp;
                               label.setText( StringUtils.abbreviate( title, TITLE_MAX_WIDTH ) );
                               if ( virtualUtil.isTooltipsVisible() )
                               {
                                   Tooltips.getCurrent()
                                           .removeTooltip( label );
                                   if ( StringUtils.length( title ) >= TITLE_MAX_WIDTH )
                                   {
                                       TooltipRunner tooltipRunner = new TooltipRunner( label, title, UI.getCurrent() );
                                       virtualUtil.startRunnerService( tooltipRunner );
                                   }
                               }
                           }
                       } );
                }
                else
                {
                    Tab newTab = createTab( id, title, icon );
                    tabs.replace( tab, newTab );
                }
            }

            if ( tabSelected )
            {
                setNewWindowFocus( id );
            }
            updated = true;
        }
        return updated;
    }

}
