/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dvirtual.xpra.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.Inflater;

import org.dussan.vaadin.dvirtual.xpra.io.brotli.PatchBrotliChunk;
import org.dussan.vaadin.dvirtual.xpra.io.inflater.ChunkInflaterInputStream;
import org.dussan.vaadin.dvirtual.xpra.io.inflater.PatchInflaterChunk;
import org.dussan.vaadin.dvirtual.xpra.io.lz4.ChunkLz4InputStream;
import org.dussan.vaadin.dvirtual.xpra.io.lz4.PatchLz4Chunk;
import org.dussan.vaadin.dvirtual.xpra.io.rencode.RencodeInputStream;

import com.dampcake.bencode.BencodeInputStream;

public class PacketReader
    implements Serializable
{

    private static final long serialVersionUID = 1037863309279212174L;

    private transient InputStream inputStream = null;

    private HeaderChunk headerChunk = null;

    public PacketReader( InputStream inputStream )
    {
        this.headerChunk = new HeaderChunk();
        this.inputStream = inputStream;
    }

    @SuppressWarnings( "resource" )
    private List<Object> readBencodedList( InputStream inputStream )
        throws IOException
    {
        BencodeInputStream bencodeInputStream = new BencodeInputStream( inputStream );
        return bencodeInputStream.readList();
    }

    @SuppressWarnings( "resource" )
    private List<Object> readRencodedList( InputStream inputStream )
        throws IOException
    {
        RencodeInputStream rencodeInputStream = new RencodeInputStream( inputStream, false );
        return rencodeInputStream.readList();
    }

    public List<Object> readList()
        throws IOException
    {
        final Map<Integer, byte[]> patches = new HashMap<>();
        boolean packetReady = false;
        while ( !packetReady )
        {
            headerChunk.readHeader( inputStream );
            if ( headerChunk.isBadHeader() )
            {
                return ( new ArrayList<>() );
            }

            if ( headerChunk.getPacketIndex() > 0 )
            {
                if ( headerChunk.isLz4Chunk() )
                {
                    // python-lz4 inserts the length of the uncompressed data as an integer value at the start of the stream
                    inputStream.read( new byte[PatchLz4Chunk.PYTHON_LZ4_LENGTH], 0, PatchLz4Chunk.PYTHON_LZ4_LENGTH );
                    byte[] patch = new PatchLz4Chunk().readPatch( inputStream, headerChunk.getPacketSize() - PatchLz4Chunk.PYTHON_LZ4_LENGTH, headerChunk.isDataCompressed() );
                    patches.put( headerChunk.getPacketIndex(), patch );
                }
                else if ( headerChunk.isBrotliChunk() )
                {
                    byte[] patch = new PatchBrotliChunk().readPatch( inputStream );
                    patches.put( headerChunk.getPacketIndex(), patch );
                }
                else
                {
                    byte[] patch = new PatchInflaterChunk().readPatch( inputStream, headerChunk.getPacketSize(), headerChunk.isDataCompressed() );
                    patches.put( headerChunk.getPacketIndex(), patch );
                }
            }
            else
            {
                packetReady = true;
            }
        }

        InputStream streamInput = inputStream;
        if ( headerChunk.isLz4Chunk() )
        {
            // python-lz4 inserts the length of the uncompressed data as an integer value at the start of the stream
            inputStream.read( new byte[PatchLz4Chunk.PYTHON_LZ4_LENGTH], 0, PatchLz4Chunk.PYTHON_LZ4_LENGTH );
            streamInput = new ChunkLz4InputStream( inputStream, headerChunk.getPacketSize() - PatchLz4Chunk.PYTHON_LZ4_LENGTH );
        }
        else if ( headerChunk.isDataCompressed() )
        {
            streamInput = new ChunkInflaterInputStream( inputStream, new Inflater(), headerChunk.getPacketSize() );
        }

        final List<Object> list;
        if ( headerChunk.isRencodeEnabled() )
        {
            list = readRencodedList( streamInput );
        }
        else
        {
            list = readBencodedList( streamInput );
        }

        if ( headerChunk.isLz4Chunk() )
        {
            ( (ChunkLz4InputStream) streamInput ).drain();
        }
        else if ( headerChunk.isDataCompressed() )
        {
            ( (ChunkInflaterInputStream) streamInput ).drain();
        }

        for ( Map.Entry<Integer, byte[]> entry : patches.entrySet() )
        {
            list.set( entry.getKey(), entry.getValue() );
        }

        return list;
    }

}
